from __future__ import annotations
from typing import List, Dict, Optional, Any
import uuid

import humps
from ..utils import selective_decamelize

from .raga import Raga
from .trajectory import Trajectory
from .chikari import Chikari
from .group import Group
from .pitch import Pitch
from .automation import get_starts
from .note_view_phrase import NoteViewPhrase


PhraseCatType = Dict[str, Dict[str, bool]]


def init_phrase_categorization() -> PhraseCatType:
    return {
        "Phrase": {
            "Mohra": False,
            "Mukra": False,
            "Asthai": False,
            "Antara": False,
            "Manjha": False,
            "Abhog": False,
            "Sanchari": False,
            "Jhala": False,
        },
        "Elaboration": {
            "Vistar": False,
            "Barhat": False,
            "Prastar": False,
            "Bol Banao": False,
            "Bol Alap": False,
            "Bol Bandt": False,
            "Behlava": False,
            "Gat-kari": False,
            "Tan (Sapat)": False,
            "Tan (Gamak)": False,
            "Laykari": False,
            "Tihai": False,
            "Chakradar": False,
        },
        "Vocal Articulation": {
            "Bol": False,
            "Non-Tom": False,
            "Tarana": False,
            "Aakar": False,
            "Sargam": False,
        },
        "Instrumental Articulation": {
            "Bol": False,
            "Non-Bol": False,
        },
        "Incidental": {
            "Talk/Conversation": False,
            "Praise ('Vah')": False,
            "Tuning": False,
            "Pause": False,
        },
    }


class Phrase:
    def __init__(self, options: Optional[Dict[str, Any]] = None) -> None:
        opts = selective_decamelize(options or {})
        
        # Parameter validation
        self._validate_parameters(opts)
        trajectories_in = opts.get('trajectories', [])
        self.start_time: Optional[float] = opts.get('start_time')
        self.raga: Optional[Raga] = opts.get('raga')
        instrumentation = opts.get('instrumentation', ['Sitar'])
        self.instrumentation: List[str] = instrumentation
        trajectory_grid_opt = opts.get('trajectory_grid')
        chikari_grid_opt = opts.get('chikari_grid')
        chikaris_in = opts.get('chikaris', {})
        groups_grid = opts.get('groups_grid')
        categorization_grid = opts.get('categorization_grid')
        unique_id = opts.get('unique_id')
        self.piece_idx = opts.get('piece_idx')
        ad_hoc_cat = opts.get('ad_hoc_categorization_grid')

        # Initialize is_section_start (optional boolean)
        self.is_section_start = opts.get('is_section_start')

        trajs: List[Trajectory] = []
        for t in trajectories_in:
            if not isinstance(t, Trajectory):
                t = Trajectory(t)  # type: ignore
            trajs.append(t)

        if trajectory_grid_opt is not None:
            self.trajectory_grid = trajectory_grid_opt
            for _ in range(len(self.trajectory_grid), len(instrumentation)):
                self.trajectory_grid.append([])
        else:
            self.trajectory_grid = [trajs]
            for _ in range(1, len(instrumentation)):
                self.trajectory_grid.append([])

        chikaris_dict: Dict[str, Chikari] = {}
        for k, v in chikaris_in.items():
            if not isinstance(v, Chikari):
                chikaris_dict[str(k)] = Chikari(v)  # type: ignore
            else:
                chikaris_dict[str(k)] = v

        if chikari_grid_opt is not None:
            self.chikari_grid = chikari_grid_opt
            for _ in range(len(self.chikari_grid), len(instrumentation)):
                self.chikari_grid.append({})
        else:
            self.chikari_grid = [chikaris_dict]
            for _ in range(1, len(instrumentation)):
                self.chikari_grid.append({})

        if len(self.trajectories) == 0:
            self.dur_tot = opts.get('dur_tot', 1)
            self.dur_array = opts.get('dur_array', [])
        else:
            self.dur_tot_from_trajectories()
            self.dur_array_from_trajectories()
            dur_tot = opts.get('dur_tot')
            if dur_tot is not None and dur_tot != self.dur_tot:
                for t in self.trajectories:
                    t.dur_tot = t.dur_tot * dur_tot / self.dur_tot
                self.dur_tot = dur_tot
            dur_array = opts.get('dur_array')
            if dur_array is not None and dur_array != self.dur_array:
                for i, t in enumerate(self.trajectories):
                    t.dur_tot = t.dur_tot * dur_array[i] / self.dur_array[i]
                self.dur_array = dur_array
                self.dur_tot_from_trajectories()

        self.assign_start_times()
        self.assign_traj_nums()

        if groups_grid is not None:
            self.groups_grid: List[List[Group]] = groups_grid
        else:
            self.groups_grid = [ [] for _ in instrumentation ]

        self.categorization_grid: List[PhraseCatType] = categorization_grid or []
        if len(self.categorization_grid) == 0:
            for _ in range(len(self.trajectory_grid)):
                self.categorization_grid.append(init_phrase_categorization())
        if self.categorization_grid[0]['Elaboration'].get('Bol Alap') is None:
            for cat in self.categorization_grid:
                cat['Elaboration']['Bol Alap'] = False

        self.ad_hoc_categorization_grid: List[str] = ad_hoc_cat or []
        self.unique_id = str(unique_id or uuid.uuid4())

    def _validate_parameters(self, opts: Dict[str, Any]) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {
            'trajectories', 'start_time', 'raga', 'instrumentation', 'trajectory_grid',
            'chikari_grid', 'chikaris', 'groups_grid', 'categorization_grid',
            'unique_id', 'piece_idx', 'ad_hoc_categorization_grid', 'dur_tot', 'dur_array',
            'is_section_start'
        }
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names with helpful suggestions
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'duration_total' or key == 'duration':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'dur_tot'?")
                elif key == 'duration_array':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'dur_array'?")
                elif key == 'start':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'start_time'?")
                elif key == 'trajectory_list' or key == 'trajs':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'trajectories'?")
                elif key == 'instruments':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'instrumentation'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        self._validate_parameter_types(opts)
        self._validate_parameter_values(opts)
    
    def _validate_parameter_types(self, opts: Dict[str, Any]) -> None:
        """Validate that all parameters have correct types."""
        if 'trajectories' in opts:
            if not isinstance(opts['trajectories'], list):
                raise TypeError(f"Parameter 'trajectories' must be a list, got {type(opts['trajectories']).__name__}")
        
        if 'start_time' in opts and opts['start_time'] is not None:
            if not isinstance(opts['start_time'], (int, float)):
                raise TypeError(f"Parameter 'start_time' must be a number, got {type(opts['start_time']).__name__}")
        
        if 'raga' in opts and opts['raga'] is not None:
            if not isinstance(opts['raga'], (Raga, dict)):
                raise TypeError(f"Parameter 'raga' must be a Raga object or dict, got {type(opts['raga']).__name__}")
        
        if 'instrumentation' in opts:
            if not isinstance(opts['instrumentation'], list):
                raise TypeError(f"Parameter 'instrumentation' must be a list, got {type(opts['instrumentation']).__name__}")
            if not all(isinstance(item, str) for item in opts['instrumentation']):
                raise TypeError("All items in 'instrumentation' must be strings")
        
        if 'trajectory_grid' in opts and opts['trajectory_grid'] is not None:
            if not isinstance(opts['trajectory_grid'], list):
                raise TypeError(f"Parameter 'trajectory_grid' must be a list, got {type(opts['trajectory_grid']).__name__}")
            if not all(isinstance(row, list) for row in opts['trajectory_grid']):
                raise TypeError("All items in 'trajectory_grid' must be lists")
        
        if 'chikari_grid' in opts and opts['chikari_grid'] is not None:
            if not isinstance(opts['chikari_grid'], list):
                raise TypeError(f"Parameter 'chikari_grid' must be a list, got {type(opts['chikari_grid']).__name__}")
            if not all(isinstance(row, dict) for row in opts['chikari_grid']):
                raise TypeError("All items in 'chikari_grid' must be dictionaries")
        
        if 'chikaris' in opts and opts['chikaris'] is not None:
            if not isinstance(opts['chikaris'], dict):
                raise TypeError(f"Parameter 'chikaris' must be a dict, got {type(opts['chikaris']).__name__}")
        
        if 'groups_grid' in opts and opts['groups_grid'] is not None:
            if not isinstance(opts['groups_grid'], list):
                raise TypeError(f"Parameter 'groups_grid' must be a list, got {type(opts['groups_grid']).__name__}")
            if not all(isinstance(row, list) for row in opts['groups_grid']):
                raise TypeError("All items in 'groups_grid' must be lists")
        
        if 'categorization_grid' in opts and opts['categorization_grid'] is not None:
            if not isinstance(opts['categorization_grid'], list):
                raise TypeError(f"Parameter 'categorization_grid' must be a list, got {type(opts['categorization_grid']).__name__}")
        
        if 'unique_id' in opts and opts['unique_id'] is not None:
            if not isinstance(opts['unique_id'], str):
                raise TypeError(f"Parameter 'unique_id' must be a string, got {type(opts['unique_id']).__name__}")
        
        if 'piece_idx' in opts and opts['piece_idx'] is not None:
            if not isinstance(opts['piece_idx'], int):
                raise TypeError(f"Parameter 'piece_idx' must be an integer, got {type(opts['piece_idx']).__name__}")
        
        if 'ad_hoc_categorization_grid' in opts and opts['ad_hoc_categorization_grid'] is not None:
            if not isinstance(opts['ad_hoc_categorization_grid'], list):
                raise TypeError(f"Parameter 'ad_hoc_categorization_grid' must be a list, got {type(opts['ad_hoc_categorization_grid']).__name__}")
            if not all(isinstance(item, str) for item in opts['ad_hoc_categorization_grid']):
                raise TypeError("All items in 'ad_hoc_categorization_grid' must be strings")
        
        if 'dur_tot' in opts and opts['dur_tot'] is not None:
            if not isinstance(opts['dur_tot'], (int, float)):
                raise TypeError(f"Parameter 'dur_tot' must be a number, got {type(opts['dur_tot']).__name__}")
        
        if 'dur_array' in opts and opts['dur_array'] is not None:
            if not isinstance(opts['dur_array'], list):
                raise TypeError(f"Parameter 'dur_array' must be a list, got {type(opts['dur_array']).__name__}")
            if not all(isinstance(item, (int, float)) for item in opts['dur_array']):
                raise TypeError("All items in 'dur_array' must be numbers")

        # Validate is_section_start
        if 'is_section_start' in opts and opts['is_section_start'] is not None:
            if not isinstance(opts['is_section_start'], bool):
                raise TypeError(f"Parameter 'is_section_start' must be a boolean, got {type(opts['is_section_start']).__name__}")

    def _validate_parameter_values(self, opts: Dict[str, Any]) -> None:
        """Validate that parameter values are in valid ranges."""
        if 'start_time' in opts and opts['start_time'] is not None:
            if opts['start_time'] < 0:
                raise ValueError(f"Parameter 'start_time' must be non-negative, got {opts['start_time']}")
        
        if 'piece_idx' in opts and opts['piece_idx'] is not None:
            if opts['piece_idx'] < 0:
                raise ValueError(f"Parameter 'piece_idx' must be non-negative, got {opts['piece_idx']}")
        
        if 'dur_tot' in opts and opts['dur_tot'] is not None:
            if opts['dur_tot'] <= 0:
                raise ValueError(f"Parameter 'dur_tot' must be positive, got {opts['dur_tot']}")
        
        if 'dur_array' in opts and opts['dur_array'] is not None:
            dur_array = opts['dur_array']
            if any(d < 0 for d in dur_array):
                raise ValueError("All values in 'dur_array' must be non-negative")
            if len(dur_array) > 0 and sum(dur_array) == 0:
                raise ValueError("'dur_array' cannot have all zero values")
        
        # Validate grid structure consistency
        if 'trajectory_grid' in opts and 'instrumentation' in opts:
            trajectory_grid = opts['trajectory_grid']
            instrumentation = opts['instrumentation']
            if len(trajectory_grid) != len(instrumentation):
                import warnings
                warnings.warn(f"trajectory_grid has {len(trajectory_grid)} tracks but instrumentation has {len(instrumentation)} instruments. "
                             "These should typically match.", UserWarning)
        
        if 'chikari_grid' in opts and 'instrumentation' in opts:
            chikari_grid = opts['chikari_grid']
            instrumentation = opts['instrumentation']
            if len(chikari_grid) != len(instrumentation):
                import warnings
                warnings.warn(f"chikari_grid has {len(chikari_grid)} tracks but instrumentation has {len(instrumentation)} instruments. "
                             "These should typically match.", UserWarning)
        
        if 'groups_grid' in opts and 'instrumentation' in opts:
            groups_grid = opts['groups_grid']
            instrumentation = opts['instrumentation']
            if len(groups_grid) != len(instrumentation):
                import warnings
                warnings.warn(f"groups_grid has {len(groups_grid)} tracks but instrumentation has {len(instrumentation)} instruments. "
                             "These should typically match.", UserWarning)
        
        # Validate categorization grid structure
        if 'categorization_grid' in opts and opts['categorization_grid'] is not None:
            for i, cat in enumerate(opts['categorization_grid']):
                if not isinstance(cat, dict):
                    raise TypeError(f"categorization_grid[{i}] must be a dictionary")
                required_categories = ['Phrase', 'Elaboration', 'Vocal Articulation', 'Instrumental Articulation', 'Incidental']
                for req_cat in required_categories:
                    if req_cat not in cat:
                        import warnings
                        warnings.warn(f"categorization_grid[{i}] missing category '{req_cat}'. "
                                     "This may cause issues with phrase categorization.", UserWarning)

    # ------------------------------------------------------------------
    def update_fundamental(self, fundamental: float) -> None:
        for traj in self.trajectories:
            traj.update_fundamental(fundamental)

    def get_groups(self, idx: int = 0) -> List[Group]:
        if idx < len(self.groups_grid) and self.groups_grid[idx] is not None:
            return self.groups_grid[idx]
        raise Exception('No groups for this index')

    def get_group_from_id(self, gid: str) -> Optional[Group]:
        for g_list in self.groups_grid:
            for g in g_list:
                if g.id == gid:
                    return g
        return None

    def assign_phrase_idx(self) -> None:
        for traj in self.trajectories:
            traj.phrase_idx = self.piece_idx

    def assign_traj_nums(self) -> None:
        for i, traj in enumerate(self.trajectories):
            traj.num = i

    def dur_tot_from_trajectories(self) -> None:
        self.dur_tot = sum(t.dur_tot for t in self.trajectories)

    def dur_array_from_trajectories(self) -> None:
        self.dur_tot_from_trajectories()
        if self.dur_tot == 0:
            self.dur_array = [0 for _ in self.trajectories]
        else:
            self.dur_array = [t.dur_tot / self.dur_tot for t in self.trajectories]

    def compute(self, x: float, log_scale: bool = False):
        if self.dur_array is None:
            raise Exception('durArray is undefined')
        if len(self.dur_array) == 0:
            return None
        starts = get_starts(self.dur_array)
        idx = 0
        for i, s in enumerate(starts):
            if x >= s:
                idx = i
            else:
                break
        inner_x = (x - starts[idx]) / self.dur_array[idx]
        traj = self.trajectories[idx]
        return traj.compute(inner_x, log_scale)

    def realign_pitches(self) -> None:
        if not self.raga:
            return
        ratios = self.raga.stratified_ratios
        for traj in self.trajectories:
            new_pitches = []
            for p in traj.pitches:
                opts = p.to_json()
                opts['ratios'] = ratios
                new_pitches.append(Pitch(opts))
            traj.pitches = new_pitches

    def assign_start_times(self) -> None:
        if self.dur_array is None:
            raise Exception('durArray is undefined')
        if self.dur_tot is None:
            raise Exception('durTot is undefined')
        starts = [s * self.dur_tot for s in get_starts(self.dur_array)]
        for traj, st in zip(self.trajectories, starts):
            traj.start_time = st

    def get_range(self) -> Dict[str, Dict[str, Any]]:
        all_pitches = [p for t in self.trajectories for p in t.pitches]
        all_pitches.sort(key=lambda p: p.frequency)
        low = all_pitches[0]
        high = all_pitches[-1]
        low_obj = {
            'frequency': low.frequency,
            'swara': low.swara,
            'oct': low.oct,
            'raised': low.raised,
            'numberedPitch': low.numbered_pitch,
        }
        high_obj = {
            'frequency': high.frequency,
            'swara': high.swara,
            'oct': high.oct,
            'raised': high.raised,
            'numberedPitch': high.numbered_pitch,
        }
        return {'min': low_obj, 'max': high_obj}

    def consolidate_silent_trajs(self) -> None:
        chain = False
        start: Optional[int] = None
        del_idxs: List[int] = []
        for i, traj in enumerate(self.trajectories):
            if traj.id == 12:
                if not chain:
                    start = i
                    chain = True
                if i == len(self.trajectories) - 1:
                    if start is None:
                        raise Exception('start is undefined')
                    extra = sum(t.dur_tot for t in self.trajectories[start+1:])
                    self.trajectories[start].dur_tot += extra
                    del_idxs.extend(range(start+1, len(self.trajectories)))
            else:
                if chain:
                    if start is None:
                        raise Exception('start is undefined')
                    extra = sum(t.dur_tot for t in self.trajectories[start+1:i])
                    self.trajectories[start].dur_tot += extra
                    del_idxs.extend(range(start+1, i))
                    chain = False
                    start = None
        new_ts: List[Trajectory] = []
        for traj in self.trajectories:
            if traj.num is None:
                raise Exception('traj.num is undefined')
            if traj.num not in del_idxs:
                new_ts.append(traj)
        self.trajectory_grid[0] = new_ts
        self.dur_array_from_trajectories()
        self.assign_start_times()
        self.assign_traj_nums()
        self.assign_phrase_idx()

    def chikaris_during_traj(self, traj: Trajectory, track: int):
        start = traj.start_time
        if start is None:
            return []
        dur = traj.dur_tot
        end = start + dur
        chikaris = self.chikari_grid[0]
        out = []
        for k, c in chikaris.items():
            time = float(k)
            if time >= start and time <= end:
                real_time = time + (self.start_time or 0)
                out.append({
                    'time': real_time,
                    'phraseTimeKey': k,
                    'phraseIdx': self.piece_idx,
                    'track': track,
                    'chikari': c,
                    'uId': c.unique_id,
                })
        return out

    # ---------------------------- properties ---------------------------
    @property
    def trajectories(self) -> List[Trajectory]:
        return self.trajectory_grid[0]

    @property
    def chikaris(self) -> Dict[str, Chikari]:
        return self.chikari_grid[0]

    @chikaris.setter
    def chikaris(self, val: Dict[str, Chikari]) -> None:
        self.chikari_grid[0] = val

    @property
    def swara(self) -> List[Dict[str, Any]]:
        swara = []
        if self.start_time is None:
            raise Exception('startTime is undefined')
        for traj in self.trajectories:
            if traj.id != 12:
                if traj.dur_array is None:
                    raise Exception('traj.durArray is undefined')
                if traj.start_time is None:
                    raise Exception('traj.startTime is undefined')
                if len(traj.dur_array) == len(traj.pitches) - 1:
                    pitches = traj.pitches[:-1]
                else:
                    pitches = traj.pitches
                for i, pitch in enumerate(pitches):
                    st = self.start_time + traj.start_time
                    time = st + get_starts(traj.dur_array)[i] * traj.dur_tot
                    swara.append({'pitch': pitch, 'time': time})
        return swara

    def all_pitches(self, repetition: bool = True) -> List[Pitch]:
        pitches: List[Pitch] = []
        for traj in self.trajectories:
            if traj.id != 12:
                pitches.extend(traj.pitches)
        if not repetition:
            out: List[Pitch] = []
            for i, p in enumerate(pitches):
                if i == 0:
                    out.append(p)
                else:
                    prev = out[-1]
                    if not (p.swara == prev.swara and p.oct == prev.oct and p.raised == prev.raised):
                        out.append(p)
            return out
        return pitches

    def first_traj_idxs(self) -> List[int]:
        idxs: List[int] = []
        ct = 0
        silent_trigger = False
        last_vowel: Optional[str] = None
        end_consonant_trigger: Optional[bool] = None
        for t_idx, traj in enumerate(self.trajectories):
            if traj.id != 12:
                c1 = ct == 0
                c2 = silent_trigger
                c3 = traj.start_consonant is not None
                c4 = end_consonant_trigger
                c5 = traj.vowel != last_vowel
                if c1 or c2 or c3 or c4 or c5:
                    idxs.append(t_idx)
                ct += 1
                end_consonant_trigger = traj.end_consonant is not None
                last_vowel = traj.vowel
            silent_trigger = traj.id == 12
        return idxs

    def traj_idx_from_time(self, time: float) -> int:
        phrase_time = time - (self.start_time or 0)
        small_offset = 1e-10
        matches = [
            traj for traj in self.trajectories
            if traj.start_time is not None and
               phrase_time >= traj.start_time - small_offset and
               phrase_time < traj.start_time + traj.dur_tot
        ]
        if not matches:
            raise Exception('No trajectory found')
        return matches[0].num  # type: ignore

    def to_json(self) -> Dict[str, Any]:
        return {
            'durTot': self.dur_tot,
            'durArray': self.dur_array,
            'chikaris': {k: c.to_json() for k, c in self.chikaris.items()},
            'raga': self.raga.to_json() if self.raga else None,
            'startTime': self.start_time,
            'trajectoryGrid': [[t.to_json() for t in row] for row in self.trajectory_grid],
            'instrumentation': self.instrumentation,
            'groupsGrid': self.groups_grid,
            'categorizationGrid': self.categorization_grid,
            'uniqueId': self.unique_id,
            'adHocCategorizationGrid': self.ad_hoc_categorization_grid,
            'isSectionStart': self.is_section_start,
        }

    @staticmethod
    def from_json(obj: Dict[str, Any]) -> 'Phrase':
        opts = selective_decamelize(obj)
        trajectory_grid = opts.get('trajectory_grid')
        if trajectory_grid is not None:
            tg = []
            for row in trajectory_grid:
                tg.append([Trajectory.from_json(t) for t in row])
            opts['trajectory_grid'] = tg
        trajectories = opts.get('trajectories')
        if trajectories is not None:
            opts['trajectories'] = [Trajectory.from_json(t) for t in trajectories]
        chikaris = opts.get('chikaris')
        if chikaris is not None:
            new_c = {}
            for k, v in chikaris.items():
                new_c[str(k)] = Chikari.from_json(v)
            opts['chikaris'] = new_c
        chikari_grid = opts.get('chikari_grid')
        if chikari_grid is not None:
            new_grid = []
            for cg in chikari_grid:
                new_obj = {}
                for k, v in cg.items():
                    new_obj[str(k)] = Chikari.from_json(v)
                new_grid.append(new_obj)
            opts['chikari_grid'] = new_grid
        raga = opts.get('raga')
        if raga is not None and not isinstance(raga, Raga):
            opts['raga'] = Raga.from_json(raga)
        return Phrase(opts)

    def to_note_view_phrase(self) -> 'NoteViewPhrase':
        pitches: List[Pitch] = []
        for traj in self.trajectories:
            if traj.id != 0:
                pitches.extend(traj.pitches)
            elif len(traj.articulations) > 0:
                pitches.extend(traj.pitches)
        return NoteViewPhrase({
            'pitches': pitches,
            'dur_tot': self.dur_tot,
            'raga': self.raga,
            'start_time': self.start_time,
        })

    def reset(self) -> None:
        self.dur_array_from_trajectories()
        self.assign_start_times()
        self.assign_phrase_idx()
        self.assign_traj_nums()





from __future__ import annotations
from typing import List, Optional, Dict, Union, Any
from datetime import datetime

from .phrase import Phrase
from .trajectory import Trajectory
from .pitch import Pitch
from .raga import Raga
from .meter import Meter
from ..enums import Instrument
from .chikari import Chikari
from .group import Group
from .automation import get_starts, get_ends
import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .section import Section

SecCatType = Dict[str, Union[Dict[str, bool], str]]


def init_sec_categorization() -> SecCatType:
    return {
        "Pre-Chiz Alap": {"Pre-Chiz Alap": False},
        "Alap": {"Alap": False, "Jor": False, "Alap-Jhala": False},
        "Composition Type": {
            "Dhrupad": False,
            "Bandish": False,
            "Thumri": False,
            "Ghazal": False,
            "Qawwali": False,
            "Dhun": False,
            "Tappa": False,
            "Bhajan": False,
            "Kirtan": False,
            "Kriti": False,
            "Masitkhani Gat": False,
            "Razakhani Gat": False,
            "Ferozkhani Gat": False,
        },
        "Comp.-section/Tempo": {
            "Ati Vilambit": False,
            "Vilambit": False,
            "Madhya": False,
            "Drut": False,
            "Ati Drut": False,
            "Jhala": False,
        },
        "Tala": {"Ektal": False, "Tintal": False, "Rupak": False},
        "Improvisation": {"Improvisation": False},
        "Other": {"Other": False},
        "Top Level": "None",
    }


# ----------------------------------------------------------------------
# Helper used outside the class
# ----------------------------------------------------------------------
def durations_of_fixed_pitches(
    trajs: List[Trajectory],
    output_type: str = "pitchNumber",
    count_type: str = "cumulative",
) -> Dict:
    pitch_durs: Dict[Any, float] = {}
    for traj in trajs:
        traj_pitch_durs = traj.durations_of_fixed_pitches({"output_type": output_type})
        if not isinstance(traj_pitch_durs, dict):
            raise SyntaxError(
                "invalid trajPitchDurs type, must be object: " + str(traj_pitch_durs)
            )
        for k, v in traj_pitch_durs.items():
            pitch_durs[k] = pitch_durs.get(k, 0.0) + float(v)

    if count_type == "proportional":
        total = sum(pitch_durs.values()) or 1.0
        for k in pitch_durs:
            pitch_durs[k] /= total
    return pitch_durs


class Piece:
    def __init__(self, options: Optional[dict] = None) -> None:
        opts = options or {}
        
        # Parameter validation
        self._validate_parameters(opts)
        raga_opt = opts.get("raga")
        if raga_opt is not None and not isinstance(raga_opt, Raga):
            raga_opt = Raga.from_json(raga_opt)
        self.raga: Raga = raga_opt or Raga()

        instrumentation = opts.get("instrumentation", [Instrument.Sitar])
        inst_list: List[Instrument] = []
        for i in instrumentation:
            if isinstance(i, Instrument):
                inst_list.append(i)
            else:
                try:
                    inst_list.append(Instrument(i))
                except ValueError:
                    inst_list.append(i)
        self.instrumentation = inst_list

        # Initialize trackTitles (after instrumentation is set)
        track_titles = opts.get('trackTitles')
        if track_titles is not None:
            self.track_titles = list(track_titles)  # Create a copy
        else:
            # Create empty strings for each instrument track
            self.track_titles = [''] * len(self.instrumentation)

        # Ensure trackTitles array matches instrumentation length
        while len(self.track_titles) < len(self.instrumentation):
            self.track_titles.append('')
        while len(self.track_titles) > len(self.instrumentation):
            self.track_titles.pop()

        self.possible_trajs: Dict[Instrument, List[int]] = {
            Instrument.Sitar: list(range(14)),
            Instrument.Vocal_M: [0, 1, 2, 3, 4, 5, 6, 12, 13],
            Instrument.Vocal_F: [0, 1, 2, 3, 4, 5, 6, 12, 13],
        }

        first_inst = self.instrumentation[0]
        self.traj_idxs = self.possible_trajs.get(first_inst, [])
        self.traj_idxs_grid = [self.possible_trajs.get(i, []) for i in self.instrumentation]

        phrase_grid = opts.get("phraseGrid")
        if phrase_grid is None:
            phrases = opts.get("phrases", [])
            phrase_grid = [phrases]
        grid: List[List[Phrase]] = []
        for row in phrase_grid:
            new_row = []
            for p in row:
                if not isinstance(p, Phrase):
                    new_row.append(Phrase.from_json(p))
                else:
                    new_row.append(p)
            grid.append(new_row)
        self.phrase_grid = grid

        self.title: str = opts.get("title", "untitled")
        self.date_created: datetime = opts.get("dateCreated", datetime.now())
        self.date_modified: datetime = opts.get("dateModified", datetime.now())
        self.location: str = opts.get("location", "Santa Cruz")
        self._id: Optional[str] = opts.get("_id")
        self.audio_id: Optional[str] = opts.get("audioID")
        self.audio_db_id: Optional[str] = opts.get("audio_DB_ID")
        self.user_id: Optional[str] = opts.get("userID")
        self.name: Optional[str] = opts.get("name")
        self.family_name: Optional[str] = opts.get("family_name")
        self.given_name: Optional[str] = opts.get("given_name")
        self.permissions: Optional[str] = opts.get("permissions")
        self.soloist: Optional[str] = opts.get("soloist")
        self.solo_instrument: Optional[str] = opts.get("soloInstrument")
        self.explicit_permissions: Dict[str, Any] = opts.get(
            "explicitPermissions",
            {"edit": [], "view": [], "publicView": True},
        )
        # Optional list of collection names this piece belongs to
        self.collections: List[str] = opts.get("collections", [])

        self.meters: List[Meter] = []
        for m in opts.get("meters", []):
            if isinstance(m, Meter):
                self.meters.append(m)
            else:
                self.meters.append(Meter.from_json(m))

        ss_grid = opts.get("sectionStartsGrid")
        if ss_grid is None:
            ss = opts.get("sectionStarts", [0])
            ss_grid = [ss]
        for _ in range(len(ss_grid), len(self.instrumentation)):
            ss_grid.append([0])
        self.section_starts_grid: List[List[float]] = [sorted(list(s)) for s in ss_grid]

        # Migrate old sectionStartsGrid to phrase-level is_section_start properties
        # This enables phrase-based section tracking while maintaining backward compatibility
        if self.section_starts_grid and self.phrase_grid:
            for inst_idx, phrases in enumerate(self.phrase_grid):
                if inst_idx < len(self.section_starts_grid):
                    starts = self.section_starts_grid[inst_idx]
                    for phrase_idx, phrase in enumerate(phrases):
                        # Convert indices to integers for comparison
                        phrase.is_section_start = phrase_idx in [int(s) for s in starts]

        sc_grid = opts.get("sectionCatGrid")
        if sc_grid is None:
            section_cat = opts.get("sectionCategorization")
            sc_grid = []
            for i, ss in enumerate(self.section_starts_grid):
                if i == 0:
                    if section_cat is not None:
                        for c in section_cat:
                            self.clean_up_section_categorization(c)
                        row = section_cat
                    else:
                        row = [init_sec_categorization() for _ in ss]
                else:
                    row = [init_sec_categorization() for _ in ss]
                sc_grid.append(row)
        self.section_cat_grid: List[List[SecCatType]] = sc_grid
        for i, ss in enumerate(self.section_starts_grid):
            while len(self.section_cat_grid) <= i:
                self.section_cat_grid.append([init_sec_categorization() for _ in ss])
            if len(self.section_cat_grid[i]) < len(ss):
                diff = len(ss) - len(self.section_cat_grid[i])
                for _ in range(diff):
                    self.section_cat_grid[i].append(init_sec_categorization())

        ad_hoc = opts.get("adHocSectionCatGrid")
        if ad_hoc is None:
            self.ad_hoc_section_cat_grid = [[[] for _ in row] for row in self.section_cat_grid]
        else:
            self.ad_hoc_section_cat_grid = [
                [ [f for f in fields if f != ""] for fields in track ]
                for track in ad_hoc
            ]
        while len(self.ad_hoc_section_cat_grid) < len(self.section_starts_grid):
            self.ad_hoc_section_cat_grid.append([[] for _ in self.ad_hoc_section_cat_grid[0]])

        self.excerpt_range = opts.get("excerptRange")
        self.assemblage_descriptors = opts.get("assemblageDescriptors", [])

        self.dur_tot: Optional[float] = opts.get("durTot")
        self.dur_array_grid: Optional[List[List[float]]] = opts.get("durArrayGrid")
        if self.dur_tot is None or self.dur_array_grid is None:
            self.dur_array_from_phrases()
        else:
            self.update_start_times()

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {
            'raga', 'instrumentation', 'phraseGrid', 'phrases', 'title', 'dateCreated',
            'dateModified', 'location', '_id', 'audioID', 'audio_DB_ID', 'userID', 'name',
            'family_name', 'given_name', 'permissions', 'soloist', 'soloInstrument',
            'explicitPermissions', 'meters', 'sectionStartsGrid', 'sectionStarts',
            'sectionCatGrid', 'sectionCategorization', 'adHocSectionCatGrid', 'excerptRange',
            'assemblageDescriptors', 'collections', 'durTot', 'durArrayGrid', 'durArray',
            'trackTitles'
        }
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names with helpful suggestions
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'phrases_grid' or key == 'phrase_grid':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'phraseGrid'?")
                elif key == 'audio_id':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'audioID'?")
                elif key == 'user_id':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'userID'?")
                elif key == 'duration_total' or key == 'dur_total':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'durTot'?")
                elif key == 'duration_array_grid':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'durArrayGrid'?")
                elif key == 'section_starts_grid':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'sectionStartsGrid'?")
                elif key == 'explicit_permissions':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'explicitPermissions'?")
                elif key == 'solo_instrument':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'soloInstrument'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        self._validate_parameter_types(opts)
        self._validate_parameter_values(opts)
    
    def _validate_parameter_types(self, opts: dict) -> None:
        """Validate that all parameters have correct types."""
        if 'raga' in opts and opts['raga'] is not None and not isinstance(opts['raga'], (Raga, dict)):
            raise TypeError(f"Parameter 'raga' must be a Raga object or dict, got {type(opts['raga']).__name__}")
        
        if 'instrumentation' in opts:
            if not isinstance(opts['instrumentation'], list):
                raise TypeError(f"Parameter 'instrumentation' must be a list, got {type(opts['instrumentation']).__name__}")
        
        if 'phraseGrid' in opts:
            if not isinstance(opts['phraseGrid'], list):
                raise TypeError(f"Parameter 'phraseGrid' must be a list, got {type(opts['phraseGrid']).__name__}")
            if not all(isinstance(row, list) for row in opts['phraseGrid']):
                raise TypeError("All items in 'phraseGrid' must be lists")
        
        if 'phrases' in opts:
            if not isinstance(opts['phrases'], list):
                raise TypeError(f"Parameter 'phrases' must be a list, got {type(opts['phrases']).__name__}")
        
        if 'title' in opts and not isinstance(opts['title'], str):
            raise TypeError(f"Parameter 'title' must be a string, got {type(opts['title']).__name__}")
        
        if 'meters' in opts:
            if not isinstance(opts['meters'], list):
                raise TypeError(f"Parameter 'meters' must be a list, got {type(opts['meters']).__name__}")
        
        if 'sectionStartsGrid' in opts:
            if not isinstance(opts['sectionStartsGrid'], list):
                raise TypeError(f"Parameter 'sectionStartsGrid' must be a list, got {type(opts['sectionStartsGrid']).__name__}")
            if not all(isinstance(row, list) for row in opts['sectionStartsGrid']):
                raise TypeError("All items in 'sectionStartsGrid' must be lists")
            if not all(all(isinstance(item, (int, float)) for item in row) for row in opts['sectionStartsGrid']):
                raise TypeError("All items in 'sectionStartsGrid' sublists must be numbers")
        
        if 'sectionStarts' in opts:
            if not isinstance(opts['sectionStarts'], list):
                raise TypeError(f"Parameter 'sectionStarts' must be a list, got {type(opts['sectionStarts']).__name__}")
            if not all(isinstance(item, (int, float)) for item in opts['sectionStarts']):
                raise TypeError("All items in 'sectionStarts' must be numbers")
        
        if 'sectionCatGrid' in opts and opts['sectionCatGrid'] is not None:
            if not isinstance(opts['sectionCatGrid'], list):
                raise TypeError(f"Parameter 'sectionCatGrid' must be a list, got {type(opts['sectionCatGrid']).__name__}")
            if not all(isinstance(row, list) for row in opts['sectionCatGrid']):
                raise TypeError("All items in 'sectionCatGrid' must be lists")
        
        if 'explicitPermissions' in opts and not isinstance(opts['explicitPermissions'], dict):
            raise TypeError(f"Parameter 'explicitPermissions' must be a dict, got {type(opts['explicitPermissions']).__name__}")
        
        if 'durTot' in opts and opts['durTot'] is not None:
            if not isinstance(opts['durTot'], (int, float)):
                raise TypeError(f"Parameter 'durTot' must be a number, got {type(opts['durTot']).__name__}")
        
        if 'durArrayGrid' in opts and opts['durArrayGrid'] is not None:
            if not isinstance(opts['durArrayGrid'], list):
                raise TypeError(f"Parameter 'durArrayGrid' must be a list, got {type(opts['durArrayGrid']).__name__}")
            if not all(isinstance(row, list) for row in opts['durArrayGrid']):
                raise TypeError("All items in 'durArrayGrid' must be lists")
            if not all(all(isinstance(item, (int, float)) for item in row) for row in opts['durArrayGrid']):
                raise TypeError("All items in 'durArrayGrid' sublists must be numbers")
        
        # Validate string parameters
        string_params = ['location', 'audioID', 'audio_DB_ID', 'userID', 'name', 
                        'family_name', 'given_name', 'permissions', 'soloist', 'soloInstrument']
        
        for param in string_params:
            if param in opts and opts[param] is not None and not isinstance(opts[param], str):
                raise TypeError(f"Parameter '{param}' must be a string, got {type(opts[param]).__name__}")
        
        # Handle _id specially since it can be string or MongoDB-style dict
        if '_id' in opts and opts['_id'] is not None:
            if not isinstance(opts['_id'], (str, dict)):
                raise TypeError(f"Parameter '_id' must be a string or dict, got {type(opts['_id']).__name__}")
        
        # Validate list parameters
        list_params = ['assemblageDescriptors', 'collections']
        for param in list_params:
            if param in opts and opts[param] is not None and not isinstance(opts[param], list):
                raise TypeError(f"Parameter '{param}' must be a list, got {type(opts[param]).__name__}")
        
        # If collections provided, ensure it is a list of strings
        if 'collections' in opts and opts['collections'] is not None:
            if not isinstance(opts['collections'], list):
                raise TypeError(f"Parameter 'collections' must be a list, got {type(opts['collections']).__name__}")
            if not all(isinstance(item, str) for item in opts['collections']):
                raise TypeError("All items in 'collections' must be strings")

        # Validate trackTitles
        if 'trackTitles' in opts and opts['trackTitles'] is not None:
            if not isinstance(opts['trackTitles'], list):
                raise TypeError(f"Parameter 'trackTitles' must be a list, got {type(opts['trackTitles']).__name__}")
            if not all(isinstance(title, str) for title in opts['trackTitles']):
                raise TypeError("All items in 'trackTitles' must be strings")

        # Handle excerptRange specially since it can be dict or list
        if 'excerptRange' in opts and opts['excerptRange'] is not None:
            if not isinstance(opts['excerptRange'], (list, dict)):
                raise TypeError(f"Parameter 'excerptRange' must be a list or dict, got {type(opts['excerptRange']).__name__}")
    
    def _validate_parameter_values(self, opts: dict) -> None:
        """Validate that parameter values are in valid ranges."""
        if 'title' in opts and opts['title'] == "":
            raise ValueError("Parameter 'title' cannot be empty")
        
        if 'durTot' in opts and opts['durTot'] is not None:
            if opts['durTot'] < 0:
                raise ValueError(f"Parameter 'durTot' must be non-negative, got {opts['durTot']}")
        
        if 'sectionStarts' in opts:
            section_starts = opts['sectionStarts']
            if any(s < 0 for s in section_starts):
                raise ValueError("All values in 'sectionStarts' must be non-negative")
        
        if 'sectionStartsGrid' in opts:
            for i, row in enumerate(opts['sectionStartsGrid']):
                if any(s < 0 for s in row):
                    raise ValueError(f"All values in 'sectionStartsGrid[{i}]' must be non-negative")
        
        if 'durArrayGrid' in opts and opts['durArrayGrid'] is not None:
            for i, row in enumerate(opts['durArrayGrid']):
                if any(d < 0 for d in row):
                    raise ValueError(f"All values in 'durArrayGrid[{i}]' must be non-negative")
                if len(row) > 0 and sum(row) == 0:
                    raise ValueError(f"'durArrayGrid[{i}]' cannot have all zero values")
        
        # Validate instrumentation contains valid instruments
        if 'instrumentation' in opts:
            for i, inst in enumerate(opts['instrumentation']):
                if not isinstance(inst, Instrument) and not isinstance(inst, str):
                    raise TypeError(f"instrumentation[{i}] must be an Instrument enum or string")
                if isinstance(inst, str):
                    try:
                        Instrument(inst)
                    except ValueError:
                        import warnings
                        warnings.warn(f"Unknown instrument name: '{inst}'. This may cause issues with instrument-specific features.", UserWarning)
        
        # Validate grid structure consistency
        if 'phraseGrid' in opts and 'instrumentation' in opts:
            phrase_grid = opts['phraseGrid']
            instrumentation = opts['instrumentation']
            if len(phrase_grid) != len(instrumentation):
                import warnings
                warnings.warn(f"phraseGrid has {len(phrase_grid)} tracks but instrumentation has {len(instrumentation)} instruments. "
                             "These should typically match.", UserWarning)
        
        # Validate explicit permissions structure
        if 'explicitPermissions' in opts:
            perms = opts['explicitPermissions']
            required_keys = {'edit', 'view', 'publicView'}
            if not required_keys.issubset(perms.keys()):
                missing = required_keys - perms.keys()
                raise ValueError(f"explicitPermissions missing required keys: {sorted(missing)}")
            
            if not isinstance(perms.get('edit'), list):
                raise TypeError("explicitPermissions['edit'] must be a list")
            if not isinstance(perms.get('view'), list):
                raise TypeError("explicitPermissions['view'] must be a list")
            if not isinstance(perms.get('publicView'), bool):
                raise TypeError("explicitPermissions['publicView'] must be a boolean")

    # ------------------------------------------------------------------
    @property
    def phrases(self) -> List[Phrase]:
        return self.phrase_grid[0]

    @phrases.setter
    def phrases(self, arr: List[Phrase]) -> None:
        self.phrase_grid[0] = arr

    @property
    def dur_array(self) -> List[float]:
        return self.dur_array_grid[0] if self.dur_array_grid else []

    @dur_array.setter
    def dur_array(self, arr: List[float]) -> None:
        if self.dur_array_grid is None:
            self.dur_array_grid = [arr]
        else:
            self.dur_array_grid[0] = arr

    @property
    def section_starts(self) -> List[float]:
        return self.section_starts_grid[0]

    @section_starts.setter
    def section_starts(self, arr: List[float]) -> None:
        self.section_starts_grid[0] = arr

    @property
    def section_categorization(self) -> List[SecCatType]:
        return self.section_cat_grid[0]

    @section_categorization.setter
    def section_categorization(self, arr: List[SecCatType]) -> None:
        self.section_cat_grid[0] = arr

    @property
    def assemblages(self) -> List["Assemblage"]:
        from .assemblage import Assemblage
        flat_phrases = [p for row in self.phrase_grid for p in row]
        return [Assemblage.from_descriptor(d, flat_phrases) for d in self.assemblage_descriptors]

    # ------------------------------------------------------------------
    def update_start_times(self) -> None:
        if not self.dur_array_grid or self.dur_tot is None:
            return
        for track, phrases in enumerate(self.phrase_grid):
            starts = [s * self.dur_tot for s in get_starts(self.dur_array_grid[track])]
            for p, st in zip(phrases, starts):
                p.start_time = st
                p.piece_idx = phrases.index(p)

    # ------------------------------------------------------------------
    def dur_tot_from_phrases(self) -> None:
        """Set ``durTot`` from contained phrases and pad shorter tracks."""
        totals = [sum(p.dur_tot for p in row) for row in self.phrase_grid]
        max_dur = max(totals) if totals else 0.0
        self.dur_tot = max_dur
        for i, total in enumerate(totals):
            if total != max_dur:
                extra = max_dur - total
                phrases = self.phrase_grid[i]
                silent = Trajectory({"id": 12, "dur_tot": extra})
                if phrases:
                    phrases[-1].trajectory_grid[0].append(silent)
                    phrases[-1].reset()
                else:
                    p = Phrase({"trajectories": [silent], "dur_tot": extra, "raga": self.raga})
                    phrases.append(p)
                    p.reset()

    def dur_array_from_phrases(self) -> None:
        """Recompute ``durArrayGrid`` removing NaN trajectories."""
        self.dur_tot_from_phrases()
        self.dur_array_grid = []
        if self.dur_tot == 0:
            self.dur_array_grid.append([])
            return
        for row in self.phrase_grid:
            arr = []
            for p in row:
                if p.dur_tot is None:
                    raise Exception("p.durTot is undefined")
                if math.isnan(p.dur_tot):
                    p.trajectory_grid[0] = [t for t in p.trajectories if not math.isnan(t.dur_tot)]
                    p.dur_tot_from_trajectories()
                arr.append(p.dur_tot / self.dur_tot)
            self.dur_array_grid.append(arr)
        self.update_start_times()

    # ------------------------------------------------------------------
    def set_dur_tot(self, dur_tot: float) -> None:
        if self.dur_tot is None:
            self.dur_tot_from_phrases()
        if self.dur_tot is None:
            self.dur_tot = dur_tot
            return
        if dur_tot < self.dur_tot:
            raise ValueError("cannot shorten duration")
        extra = dur_tot - self.dur_tot
        if extra > 0 and self.phrases:
            last_phrase = self.phrases[-1]
            if last_phrase.trajectories and last_phrase.trajectories[-1].id == 12:
                last_phrase.trajectories[-1].dur_tot += extra
            else:
                silent = Trajectory({"id": 12, "dur_tot": extra})
                last_phrase.trajectory_grid[0].append(silent)
            last_phrase.reset()
        self.dur_tot = dur_tot
        self.dur_array_from_phrases()

    def fill_remaining_duration(self, target_duration: float, track: int = 0) -> None:
        """Add a silent trajectory to fill the remaining duration to reach target_duration.
        
        Args:
            target_duration: The desired total duration for the piece
            track: Which instrument track to add the silence to (default: 0)
        """
        if self.dur_tot is None:
            self.dur_tot_from_phrases()
        
        current_duration = sum(p.dur_tot for p in self.phrase_grid[track])
        
        if target_duration <= current_duration:
            return  # Already at or exceeding target duration
            
        remaining = target_duration - current_duration
        
        # Create silent trajectory to fill the remaining duration
        silent_traj = Trajectory({
            'id': 12,
            'dur_tot': remaining,
            'fundID12': self.raga.fundamental
        })
        
        # Add as a new phrase or append to existing phrase
        if self.phrase_grid[track]:
            # Add to last phrase
            last_phrase = self.phrase_grid[track][-1]
            last_phrase.trajectory_grid[0].append(silent_traj)
            last_phrase.reset()
        else:
            # Create new phrase with silent trajectory
            silent_phrase = Phrase({
                'trajectories': [silent_traj],
                'dur_tot': remaining,
                'raga': self.raga
            })
            self.phrase_grid[track].append(silent_phrase)
        
        # Update piece-level durations
        self.dur_array_from_phrases()

    def add_trajectory(
        self,
        trajectory_data: Union[Trajectory, Dict[str, Any]], 
        inst_track: int,
        start_time: float
    ) -> bool:
        """Add a trajectory to the piece by replacing part of a silent trajectory.
        
        Args:
            trajectory_data: Either a Trajectory object or dict with trajectory parameters
            inst_track: Index of the instrument track to add the trajectory to
            start_time: Start time in the piece (in seconds)
            
        Returns:
            bool: True if trajectory was successfully added, False otherwise
        """
        # Validate inputs
        if not self._validate_trajectory_addition_inputs(trajectory_data, inst_track, start_time):
            return False
        
        # Prepare trajectory with piece context
        trajectory = self._prepare_trajectory_for_addition(trajectory_data, inst_track)
        
        # Validate all requirements
        if not self._validate_trajectory_addition_requirements(trajectory, inst_track, start_time):
            return False
        
        # Find target phrase and silent trajectory
        phrase, silent_traj_idx = self._find_target_location(inst_track, start_time)
        if phrase is None or silent_traj_idx is None:
            return False
        
        # Apply replacement logic
        self._replace_silent_trajectory(phrase, silent_traj_idx, trajectory, start_time)
        
        # Update piece state
        phrase.reset()
        return True

    def _validate_trajectory_addition_inputs(
        self, 
        trajectory_data: Union[Trajectory, Dict[str, Any]], 
        inst_track: int, 
        start_time: float
    ) -> bool:
        """Validate basic inputs for trajectory addition."""
        if not isinstance(inst_track, int) or inst_track < 0:
            return False
        if not isinstance(start_time, (int, float)) or start_time < 0:
            return False
        if inst_track >= len(self.phrase_grid):
            return False
        return True

    def _prepare_trajectory_for_addition(
        self, 
        trajectory_data: Union[Trajectory, Dict[str, Any]], 
        inst_track: int
    ) -> Trajectory:
        """Convert input to Trajectory object with piece context."""
        if isinstance(trajectory_data, Trajectory):
            # Even for existing Trajectory objects, recreate with piece's raga/context
            traj_dict = trajectory_data.to_json()
        else:
            traj_dict = dict(trajectory_data)
        
        # Ensure trajectory has piece context
        traj_dict['instrumentation'] = self.instrumentation[inst_track]
        
        # Update pitches to use piece's raga fundamental  
        if 'pitches' in traj_dict:
            updated_pitches = []
            for pitch_data in traj_dict['pitches']:
                if isinstance(pitch_data, dict):
                    pitch_data['fundamental'] = self.raga.fundamental
                    pitch_data['raga'] = self.raga.to_json()
                    updated_pitches.append(Pitch(pitch_data))
                else:
                    updated_pitches.append(pitch_data)
            traj_dict['pitches'] = updated_pitches
        
        # Set fundID12 for silent trajectories
        if traj_dict.get('id') == 12:
            traj_dict['fundID12'] = self.raga.fundamental
        
        trajectory = Trajectory(traj_dict)
        return trajectory

    def _validate_trajectory_addition_requirements(
        self, 
        trajectory: Trajectory, 
        inst_track: int, 
        start_time: float
    ) -> bool:
        """Validate all 5 requirements for trajectory addition."""
        end_time = start_time + trajectory.dur_tot
        
        # 1. Time range must be within piece duration
        if self.dur_tot is None:
            self.dur_tot_from_phrases()
        if end_time > self.dur_tot:
            return False
        
        # 2. Track must be valid (already validated in inputs)
        
        # 3. Time range must be within a single phrase
        phrase_idx = self.phrase_idx_from_time(start_time, inst_track)
        phrase = self.phrase_grid[inst_track][phrase_idx]
        phrase_start = phrase.start_time or 0
        phrase_end = phrase_start + phrase.dur_tot
        
        if end_time > phrase_end:
            return False
        
        # 4. Time range must be within a single trajectory
        target_traj = self.traj_from_time(start_time, inst_track)
        if target_traj is None:
            return False
        
        # Find trajectory's absolute start time
        traj_abs_start = phrase_start + (target_traj.start_time or 0)
        traj_abs_end = traj_abs_start + target_traj.dur_tot
        
        if start_time < traj_abs_start or end_time > traj_abs_end:
            return False
        
        # 5. Target trajectory must be silent (id == 12)
        if target_traj.id != 12:
            return False
        
        return True

    def _find_target_location(self, inst_track: int, start_time: float) -> tuple[Optional[Phrase], Optional[int]]:
        """Find the phrase and trajectory index for the target location."""
        phrase_idx = self.phrase_idx_from_time(start_time, inst_track)
        phrase = self.phrase_grid[inst_track][phrase_idx]
        
        # Find the trajectory index within the phrase
        phrase_start = phrase.start_time or 0
        relative_start_time = start_time - phrase_start
        
        current_time = 0.0
        for i, traj in enumerate(phrase.trajectories):
            traj_start = current_time
            traj_end = current_time + traj.dur_tot
            
            if traj_start <= relative_start_time < traj_end:
                return phrase, i
            
            current_time = traj_end
        
        return None, None

    def _replace_silent_trajectory(
        self, 
        phrase: Phrase, 
        silent_traj_idx: int, 
        new_trajectory: Trajectory, 
        start_time: float
    ) -> None:
        """Replace part or all of a silent trajectory with the new trajectory."""
        silent_traj = phrase.trajectories[silent_traj_idx]
        phrase_start = phrase.start_time or 0
        silent_abs_start = phrase_start + (silent_traj.start_time or 0)
        
        # Calculate relative times within the silent trajectory
        rel_start = start_time - silent_abs_start
        rel_end = rel_start + new_trajectory.dur_tot
        
        trajs = phrase.trajectories
        
        # Determine which case we're in (from NewTrajEmit logic)
        starts_equal = abs(rel_start) < 1e-10  # essentially zero
        ends_equal = abs(rel_end - silent_traj.dur_tot) < 1e-10
        
        if starts_equal and ends_equal:
            # Case 1: Replace entire silent trajectory
            trajs[silent_traj_idx] = new_trajectory
        elif starts_equal:
            # Case 2: Replace left side of silent trajectory
            silent_traj.dur_tot = silent_traj.dur_tot - new_trajectory.dur_tot
            trajs.insert(silent_traj_idx, new_trajectory)
        elif ends_equal:
            # Case 3: Replace right side of silent trajectory
            silent_traj.dur_tot = silent_traj.dur_tot - new_trajectory.dur_tot
            trajs.insert(silent_traj_idx + 1, new_trajectory)
        else:
            # Case 4: Replace internal portion of silent trajectory
            first_dur = rel_start
            last_dur = silent_traj.dur_tot - rel_end
            
            # Shorten the original silent trajectory to the first part
            silent_traj.dur_tot = first_dur
            
            # Create new silent trajectory for the remaining part
            last_silent_traj = Trajectory({
                'id': 12,
                'pitches': [],
                'dur_tot': last_dur,
                'fundID12': self.raga.fundamental,
                'instrumentation': self.instrumentation[self.phrase_grid.index(next(row for row in self.phrase_grid if phrase in row))]
            })
            
            # Insert new trajectory and remaining silent trajectory
            trajs.insert(silent_traj_idx + 1, new_trajectory)
            trajs.insert(silent_traj_idx + 2, last_silent_traj)

    def realign_pitches(self) -> None:
        for phrases in self.phrase_grid:
            for p in phrases:
                p.realign_pitches()

    def update_fundamental(self, fundamental: float) -> None:
        self.raga.fundamental = fundamental
        for phrases in self.phrase_grid:
            for p in phrases:
                p.update_fundamental(fundamental)

    def put_raga_in_phrase(self) -> None:
        for phrases in self.phrase_grid:
            for p in phrases:
                p.raga = self.raga

    # ------------------------------------------------------------------
    @property
    def sections_grid(self) -> List[List["Section"]]:
        from .section import Section
        grid: List[List["Section"]] = []
        for i, starts in enumerate(self.section_starts_grid):
            sections: List["Section"] = []
            for j, s in enumerate(starts):
                if j == len(starts) - 1:
                    slice_phrases = self.phrase_grid[i][s:]
                else:
                    slice_phrases = self.phrase_grid[i][s:starts[j + 1]]
                sections.append(
                    Section(
                        {
                            "phrases": slice_phrases,
                            "categorization": self.section_cat_grid[i][j],
                            "ad_hoc_categorization": self.ad_hoc_section_cat_grid[i][j],
                        }
                    )
                )
            grid.append(sections)
        return grid

    @property
    def sections(self) -> List["Section"]:
        return self.sections_grid[0]

    def add_meter(self, meter: Meter) -> None:
        for m in self.meters:
            dur_m = m.cycle_dur
            dur_new = meter.cycle_dur
            c1 = m.start_time <= meter.start_time < m.start_time + dur_m
            c2 = m.start_time < meter.start_time + dur_new <= m.start_time + dur_m
            c3 = meter.start_time <= m.start_time and meter.start_time + dur_new >= m.start_time + dur_m
            if c1 or c2 or c3:
                raise ValueError("meters overlap")
        self.meters.append(meter)

    def remove_meter(self, meter: Meter) -> None:
        if meter in self.meters:
            self.meters.remove(meter)

    # ------------------------------------------------------------------
    def all_trajectories(self, inst: int = 0) -> List[Trajectory]:
        trajs: List[Trajectory] = []
        for p in self.phrase_grid[inst]:
            trajs.extend(p.trajectories)
        return trajs

    # ------------------------------------------------------------------
    def track_from_traj(self, traj: Trajectory) -> int:
        for i in range(len(self.instrumentation)):
            if traj in self.all_trajectories(i):
                return i
        raise ValueError("Trajectory not found")

    def track_from_traj_uid(self, traj_uid: str) -> int:
        for i in range(len(self.instrumentation)):
            for t in self.all_trajectories(i):
                if t.unique_id == traj_uid:
                    return i
        raise ValueError("Trajectory not found")

    def phrase_from_uid(self, uid: str) -> Phrase:
        for track in self.phrase_grid:
            for p in track:
                if p.unique_id == uid:
                    return p
        raise ValueError("Phrase not found")

    def track_from_phrase_uid(self, uid: str) -> int:
        for i, track in enumerate(self.phrase_grid):
            if any(p.unique_id == uid for p in track):
                return i
        raise ValueError("Phrase not found")

    def traj_from_uid(self, uid: str, track: int = 0) -> Trajectory:
        for t in self.all_trajectories(track):
            if t.unique_id == uid:
                return t
        raise ValueError("Trajectory not found")

    def traj_from_time(self, time: float, track: int = 0) -> Optional[Trajectory]:
        trajs = self.all_trajectories(track)
        if not trajs:
            return None
        starts = self.traj_start_times(track)
        end_times = [s + t.dur_tot for s, t in zip(starts, trajs)]
        idx = -1
        for i, s in enumerate(starts):
            if time >= s:
                idx = i
        if idx == -1:
            return trajs[0]
        if time < end_times[idx]:
            return trajs[idx]
        if idx + 1 < len(trajs):
            return trajs[idx + 1]
        return None

    def phrase_from_time(self, time: float, track: int = 0) -> Phrase:
        starts = self.dur_starts(track)
        idx = 0
        for i, s in enumerate(starts):
            if time >= s:
                idx = i
        return self.phrase_grid[track][idx]

    def phrase_idx_from_time(self, time: float, track: int = 0) -> int:
        starts = self.dur_starts(track)
        idx = 0
        for i, s in enumerate(starts):
            if time >= s:
                idx = i
        return idx

    def all_groups(self, instrument_idx: int = 0) -> List["Group"]:
        groups: List["Group"] = []
        for p in self.phrase_grid[instrument_idx]:
            for g_list in p.groups_grid:
                groups.extend(g_list)
        return groups

    def p_idx_from_group(self, g: "Group") -> int:
        for i, p in enumerate(self.phrase_grid[0]):
            for group_list in p.groups_grid:
                if g in group_list:
                    return i
        return -1

    def all_display_vowels(self, inst: int = 0) -> List[Any]:
        vocal = [Instrument.Vocal_M, Instrument.Vocal_F]
        if self.instrumentation[inst] not in vocal:
            raise Exception("instrumentation is not vocal")
        # placeholder minimal return
        return []

    def s_idx_from_p_idx(self, p_idx: int, inst: int = 0) -> int:
        ss = self.section_starts_grid[inst]
        s_idx = len(ss) - 1
        for i, s in enumerate(ss):
            if p_idx < s:
                s_idx = i - 1
                break
        if s_idx < 0:
            s_idx = 0
        return s_idx

    def durations_of_fixed_pitches(
        self, inst: int = 0, output_type: str = "pitchNumber"
    ) -> Dict:
        trajs = self.all_trajectories(inst)
        return durations_of_fixed_pitches(trajs, output_type)

    def proportions_of_fixed_pitches(
        self, inst: int = 0, output_type: str = "pitchNumber"
    ) -> Dict:
        return durations_of_fixed_pitches(trajs=self.all_trajectories(inst), output_type=output_type, count_type="proportional")

    # ------------------------------------------------------------------
    def chikari_freqs(self, inst_idx: int) -> List[float]:
        phrases = self.phrase_grid[inst_idx]
        for p in phrases:
            if p.chikaris:
                chikari = list(p.chikaris.values())[0]
                return [c.frequency for c in chikari.pitches[:2]]
        f = self.raga.fundamental
        return [f * 2, f * 4]

    # ------------------------------------------------------------------
    def dur_starts(self, track: int = 0) -> List[float]:
        if self.dur_array_grid is None:
            raise Exception("durArray is undefined")
        if self.dur_tot is None:
            raise Exception("durTot is undefined")
        return get_starts([d * self.dur_tot for d in self.dur_array_grid[track]])

    def traj_start_times(self, inst: int = 0) -> List[float]:
        trajs = self.all_trajectories(inst)
        times = [0.0]
        for t in trajs[:-1]:
            times.append(times[-1] + t.dur_tot)
        return times

    def all_pitches(self, repetition: bool = True, pitch_number: bool = False, track: int = 0) -> List[Any]:
        pitches: List[Any] = []
        for p in self.phrase_grid[track]:
            pitches.extend(p.all_pitches(True))
        if not repetition:
            out: List[Any] = []
            for i, pitch in enumerate(pitches):
                if isinstance(pitch, (int, float)):
                    raise ValueError("pitch is a number")
                if i == 0:
                    out.append(pitch)
                else:
                    prev = out[-1]
                    if isinstance(prev, (int, float)):
                        raise ValueError("lastP is a number")
                    if not (pitch.swara == prev.swara and pitch.oct == prev.oct and pitch.raised == prev.raised):
                        out.append(pitch)
            pitches = out
        if pitch_number:
            nums: List[Any] = []
            for p in pitches:
                if isinstance(p, (int, float)):
                    nums.append(p)
                else:
                    nums.append(p.numbered_pitch)
            return nums
        return pitches

    @property
    def highest_pitch_number(self) -> float:
        return max(self.all_pitches(pitch_number=True))

    @property
    def lowest_pitch_number(self) -> float:
        return min(self.all_pitches(pitch_number=True))

    def most_recent_traj(self, time: float, inst: int = 0) -> Trajectory:
        trajs = self.all_trajectories(inst)
        end_times: List[float] = []
        for t in trajs:
            phrase = next((p for p in self.phrase_grid[inst] if t in p.trajectories), None)
            if phrase is None:
                continue
            end_times.append((phrase.start_time or 0) + (t.start_time or 0) + t.dur_tot)
        latest = max([et for et in end_times if et <= time], default=-float("inf"))
        idx = end_times.index(latest)
        return trajs[idx]

    # ------------------------------------------------------------------
    def chunked_trajs(self, inst: int = 0, duration: float = 30) -> List[List[Trajectory]]:
        trajs = self.all_trajectories(inst)
        durs = [t.dur_tot for t in trajs]
        starts = get_starts(durs)
        ends = get_ends(durs)
        chunks: List[List[Trajectory]] = []
        if self.dur_tot is None:
            self.dur_tot_from_phrases()
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            def f1(st: float) -> bool:
                return st >= i and st < i + duration

            def f2(et: float) -> bool:
                return et > i and et <= i + duration

            def f3(st: float, et: float) -> bool:
                return st < i and et > i + duration

            chunk = [trajs[j] for j in range(len(trajs)) if f1(starts[j]) or f2(ends[j]) or f3(starts[j], ends[j])]
            chunks.append(chunk)
            i += duration
        return chunks

    def all_display_bols(self, inst: int = 0) -> List[Dict[str, Any]]:
        trajs = self.all_trajectories(inst)
        starts = self.traj_start_times(inst)
        idxs: List[int] = []
        bol_trajs: List[Trajectory] = []
        for idx, t in enumerate(trajs):
            art = t.articulations.get("0.00")
            if art and art.name == "pluck":
                idxs.append(idx)
                bol_trajs.append(t)
        bols = []
        for i, t in enumerate(bol_trajs):
            art = t.articulations["0.00"]
            bols.append({
                "time": starts[idxs[i]],
                "bol": getattr(art, "stroke_nickname", None),
                "uId": t.unique_id,
                "logFreq": t.log_freqs[0],
                "track": inst,
            })
        return bols

    def all_display_sargam(self, inst: int = 0) -> List[Dict[str, Any]]:
        trajs = self.all_trajectories(inst)
        starts = self.traj_start_times(inst)
        sargams: List[Dict[str, Any]] = []
        last_pitch: Dict[str, Optional[float]] = {"logFreq": None, "time": None}
        for i, t in enumerate(trajs):
            if t.id != 12:
                sub_durs = [d * t.dur_tot for d in (t.dur_array or [])]
                time_pts = get_starts(sub_durs)
                time_pts.append(t.dur_tot)
                time_pts = [tp + starts[i] for tp in time_pts]
                for tp_idx, tp in enumerate(time_pts):
                    log_freq = t.log_freqs[tp_idx] if tp_idx < len(t.log_freqs) else t.log_freqs[tp_idx - 1]
                    c_lf = last_pitch["logFreq"] == log_freq
                    c_t = last_pitch["time"] == tp
                    if not (c_lf or (c_lf and c_t)):
                        pitch = t.pitches[min(tp_idx, len(t.pitches)-1)]
                        sargams.append({
                            "logFreq": log_freq,
                            "sargam": pitch.sargam_letter,
                            "time": tp,
                            "uId": t.unique_id,
                            "track": inst,
                            "solfege": pitch.solfege_letter,
                            "pitchClass": str(pitch.chroma),
                            "westernPitch": pitch.western_pitch,
                        })
                    last_pitch = {"logFreq": log_freq, "time": tp}
        phrase_divs = [p.start_time + p.dur_tot for p in self.phrase_grid[inst]]
        pwr = 10 ** 5
        rounded_pds = [round(pd * pwr) / pwr for pd in phrase_divs]
        for s_idx, s in enumerate(sargams):
            pos = 1
            last_higher = True
            next_higher = True
            if s_idx != 0 and s_idx != len(sargams) - 1:
                last_s = sargams[s_idx - 1]
                next_s = sargams[s_idx + 1]
                last_higher = last_s["logFreq"] > s["logFreq"]
                next_higher = next_s["logFreq"] > s["logFreq"]
            if last_higher and next_higher:
                pos = 0
            elif not last_higher and not next_higher:
                pos = 1
            elif last_higher and not next_higher:
                pos = 3
            elif not last_higher and next_higher:
                pos = 2
            if round(s["time"] * pwr) / pwr in rounded_pds:
                pos = 5 if next_higher else 4
            s["pos"] = pos
        return sargams

    def all_phrase_divs(self, inst: int = 0) -> List[Dict[str, Any]]:
        objs: List[Dict[str, Any]] = []
        for p_idx, p in enumerate(self.phrase_grid[inst]):
            if p_idx != 0:
                objs.append({
                    "time": p.start_time,
                    "type": "section" if p_idx in self.section_starts_grid[inst] else "phrase",
                    "idx": p_idx,
                    "track": inst,
                    "uId": p.unique_id,
                })
        return objs

    def all_display_vowels(self, inst: int = 0) -> List[Dict[str, Any]]:
        vocal = [Instrument.Vocal_M, Instrument.Vocal_F]
        if self.instrumentation[inst] not in vocal:
            raise Exception("instrumentation is not vocal")
        display_vowels: List[Dict[str, Any]] = []
        for phrase in self.phrase_grid[inst]:
            first_idxs = phrase.first_traj_idxs()
            phrase_start = phrase.start_time or 0
            for t_idx in first_idxs:
                traj = phrase.trajectories[t_idx]
                time = phrase_start + (traj.start_time or 0)
                log_freq = traj.log_freqs[0]
                with_c = traj.start_consonant is not None
                art = traj.articulations.get("0.00") if with_c else None
                ipa_text = (art.ipa if art else "") + (traj.vowel_ipa or "")
                devanagari_text = (art.hindi if art else "") + (traj.vowel_hindi or "")
                english_text = (art.eng_trans if art else "") + (traj.vowel_eng_trans or "")
                display_vowels.append({
                    "time": time,
                    "logFreq": log_freq,
                    "ipaText": ipa_text,
                    "devanagariText": devanagari_text,
                    "englishText": english_text,
                    "uId": traj.unique_id,
                })
        return display_vowels

    def all_display_ending_consonants(self, inst: int = 0) -> List[Dict[str, Any]]:
        display: List[Dict[str, Any]] = []
        trajs = self.all_trajectories(inst)
        for t in trajs:
            if t.end_consonant is not None:
                phrase = next((p for p in self.phrase_grid[inst] if t in p.trajectories), None)
                phrase_start = phrase.start_time if phrase else 0
                time = phrase_start + (t.start_time or 0) + t.dur_tot
                log_freq = t.log_freqs[-1]
                art = t.articulations.get("1.00")
                display.append({
                    "time": time,
                    "logFreq": log_freq,
                    "ipaText": art.ipa if art else None,
                    "devanagariText": art.hindi if art else None,
                    "englishText": art.eng_trans if art else None,
                    "uId": t.unique_id,
                })
        return display

    def all_display_chikaris(self, inst: int = 0) -> List[Dict[str, Any]]:
        display: List[Dict[str, Any]] = []
        for p in self.phrase_grid[inst]:
            for k, chikari in p.chikaris.items():
                time = p.start_time + float(k)
                display.append({
                    "time": time,
                    "phraseTimeKey": k,
                    "phraseIdx": p.piece_idx,
                    "track": inst,
                    "chikari": chikari,
                    "uId": chikari.unique_id,
                })
        return display

    def chunked_display_chikaris(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        display = self.all_display_chikaris(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [c for c in display if c["time"] >= i and c["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_display_consonants(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        display = self.all_display_ending_consonants(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [c for c in display if c["time"] >= i and c["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_display_vowels(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        display = self.all_display_vowels(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [v for v in display if v["time"] >= i and v["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_display_sargam(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        display = self.all_display_sargam(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [s for s in display if s["time"] >= i and s["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_display_bols(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        display = self.all_display_bols(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [b for b in display if b["time"] >= i and b["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_phrase_divs(self, inst: int = 0, duration: float = 30) -> List[List[Dict[str, Any]]]:
        phrase_divs = self.all_phrase_divs(inst)
        chunks: List[List[Dict[str, Any]]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [pd for pd in phrase_divs if pd["time"] >= i and pd["time"] < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def chunked_meters(self, duration: float = 30) -> List[List[Meter]]:
        chunks: List[List[Meter]] = []
        dur_tot = self.dur_tot or 0.0
        i = 0.0
        while i < dur_tot:
            chunk = [m for m in self.meters if m.start_time >= i and m.start_time < i + duration]
            chunks.append(chunk)
            i += duration
        return chunks

    def pulse_from_id(self, id: str):
        all_pulses = [p for m in self.meters for p in m.all_pulses]
        for pulse in all_pulses:
            if pulse.unique_id == id:
                return pulse
        return None

    # ------------------------------------------------------------------
    def clean_up_section_categorization(self, c: SecCatType) -> None:
        if "Improvisation" not in c:
            c["Improvisation"] = {"Improvisation": False}
        if "Other" not in c:
            c["Other"] = {"Other": False}
        if "Comp.-section/Tempo" not in c and "Composition-section/Tempo" in c:
            c["Comp.-section/Tempo"] = c["Composition-section/Tempo"]
            del c["Composition-section/Tempo"]
        top_level = c.get("Top Level")
        if not top_level or top_level == "None":
            if any(c["Pre-Chiz Alap"].values()):
                c["Top Level"] = "Pre-Chiz Alap"
            elif any(c["Alap"].values()):
                c["Top Level"] = "Alap"
            elif any(c["Composition Type"].values()):
                c["Top Level"] = "Composition"
            elif any(c["Comp.-section/Tempo"].values()):
                c["Top Level"] = "Composition"
            elif any(c["Improvisation"].values()):
                c["Top Level"] = "Improvisation"
            elif any(c["Other"].values()):
                c["Top Level"] = "Other"
            else:
                c["Top Level"] = "None"

    # ------------------------------------------------------------------
    def to_json(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {
            "raga": self.raga.to_json(),
            "phraseGrid": [[p.to_json() for p in row] for row in self.phrase_grid],
            "instrumentation": [i.value if isinstance(i, Instrument) else i for i in self.instrumentation],
            "trackTitles": self.track_titles,
            "durTot": self.dur_tot,
            "durArrayGrid": self.dur_array_grid,
            "meters": [m.to_json() for m in self.meters],
            "title": self.title,
            "dateCreated": self.date_created.isoformat(),
            "dateModified": self.date_modified.isoformat(),
            "location": self.location,
            "_id": self._id,
            "audioID": self.audio_id,
            "userID": self.user_id,
            "permissions": self.permissions,
            "name": self.name,
            "family_name": self.family_name,
            "given_name": self.given_name,
            "sectionStartsGrid": self.section_starts_grid,
            "sectionCatGrid": self.section_cat_grid,
            "explicitPermissions": self.explicit_permissions,
            "soloist": self.soloist,
            "soloInstrument": self.solo_instrument,
            "excerptRange": self.excerpt_range,
            "adHocSectionCatGrid": self.ad_hoc_section_cat_grid,
            "assemblageDescriptors": self.assemblage_descriptors,
            "collections": self.collections,
        }
        # drop None values so they serialize as undefined (omitted) rather than null
        return {k: v for k, v in data.items() if v is not None}

    @staticmethod
    def from_json(obj: Dict[str, Any]) -> "Piece":
        new_obj = dict(obj)
        if "raga" in new_obj:
            new_obj["raga"] = Raga.from_json(new_obj["raga"])
        if "phraseGrid" in new_obj:
            pg = []
            for row in new_obj["phraseGrid"]:
                phrase_row = [Phrase.from_json(p) for p in row]
                pg.append(phrase_row)
            # reconstruct groups so they reference existing trajectories
            for row in pg:
                for phrase in row:
                    new_groups: List[List[Group]] = []
                    for g_list in phrase.groups_grid:
                        rebuilt: List[Group] = []
                        for g in g_list:
                            data = g if isinstance(g, dict) else g.to_json()
                            trajs = []
                            for t in data.get("trajectories", []):
                                traj = t if isinstance(t, Trajectory) else Trajectory.from_json(t)
                                num = traj.num
                                if num is None or num >= len(phrase.trajectory_grid[0]):
                                    continue
                                real_traj = phrase.trajectory_grid[0][num]
                                art = real_traj.articulations.get("0.00") or real_traj.articulations.get("0")
                                if art and art.name == "slide":
                                    art.name = "pluck"
                                trajs.append(real_traj)
                            rebuilt.append(Group({"trajectories": trajs, "id": data.get("id")}))
                        new_groups.append(rebuilt)
                    phrase.groups_grid = new_groups
            new_obj["phraseGrid"] = pg
        if "meters" in new_obj:
            new_obj["meters"] = [Meter.from_json(m) for m in new_obj["meters"]]
        if "dateCreated" in new_obj:
            dc = new_obj["dateCreated"]
            if isinstance(dc, dict) and "$date" in dc:
                dc = dc["$date"]
            new_obj["dateCreated"] = datetime.fromisoformat(str(dc).replace('Z',''))
        if "dateModified" in new_obj:
            dm = new_obj["dateModified"]
            if isinstance(dm, dict) and "$date" in dm:
                dm = dm["$date"]
            new_obj["dateModified"] = datetime.fromisoformat(str(dm).replace('Z',''))

        piece = Piece(new_obj)

        # reconnect groups to actual trajectories
        for phrases in piece.phrase_grid:
            for phrase in phrases:
                new_group_grid: List[List[Group]] = []
                for group_list in phrase.groups_grid:
                    rebuilt: List[Group] = []
                    for g in group_list:
                        if not isinstance(g, Group):
                            g = Group.from_json(g)
                        new_trajs: List[Trajectory] = []
                        for t in g.trajectories:
                            if t.num is None:
                                raise Exception("traj.num is undefined")
                            new_trajs.append(phrase.trajectory_grid[0][t.num])
                        rebuilt.append(Group({"trajectories": new_trajs, "id": g.id}))
                    new_group_grid.append(rebuilt)
                phrase.groups_grid = new_group_grid

                for traj in phrase.trajectories:
                    art = traj.articulations.get("0.00")
                    if art and art.name == "slide":
                        art.name = "pluck"
                phrase.consolidate_silent_trajs()

        piece.dur_array_from_phrases()
        piece.section_starts_grid = [sorted(set(arr)) for arr in piece.section_starts_grid]

        return piece



import os
import sys
sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.chikari import Chikari
from idtap.classes.pitch import Pitch

# Test mirrors src/js/tests/chikari.test.ts

def test_chikari_serialization():
    pitches = [Pitch({'swara': 's', 'oct': 1}), Pitch({'swara': 'p'})]
    c = Chikari({'pitches': pitches, 'fundamental': 440})
    assert isinstance(c.unique_id, str)
    json_obj = c.to_json()
    copy = Chikari.from_json(json_obj)
    assert copy.to_json() == json_obj



from __future__ import annotations
import math
import uuid
from typing import List, Dict, Optional, Callable, TypedDict

import humps

from .pitch import Pitch
from .articulation import Articulation
from .automation import Automation, get_starts
from ..enums import Instrument


class VibObjType(TypedDict, total=False):
    periods: int
    vert_offset: float
    init_up: bool
    extent: float


class Trajectory:
    def __init__(self, options: Optional[dict] = None) -> None:
        opts = humps.decamelize(options or {})
        
        # Parameter validation
        self._validate_parameters(opts)
        self.names = [
            'Fixed',
            'Bend: Simple',
            'Bend: Sloped Start',
            'Bend: Sloped End',
            'Bend: Ladle',
            'Bend: Reverse Ladle',
            'Bend: Simple Multiple',
            'Krintin',
            'Krintin Slide',
            'Krintin Slide Hammer',
            'Dense Krintin Slide Hammer',
            'Slide',
            'Silent',
            'Vibrato'
        ]

        id_val = opts.get('id', 0)
        if not isinstance(id_val, int):
            raise SyntaxError(f'invalid id type, must be int: {id_val}')
        self.id: int = id_val

        pitches_in = opts.get('pitches', [Pitch()])
        if not isinstance(pitches_in, list) or not all(isinstance(p, Pitch) for p in pitches_in):
            raise SyntaxError('invalid pitches type, must be array of Pitch: ' + str(pitches_in))
        self.pitches: List[Pitch] = pitches_in

        dur_tot = opts.get('dur_tot', 1.0)
        if not isinstance(dur_tot, (int, float)):
            raise SyntaxError(f'invalid durTot type, must be number: {dur_tot}')
        self.dur_tot: float = float(dur_tot)

        self.dur_array: Optional[List[float]] = opts.get('dur_array')

        slope = opts.get('slope')
        if slope is None:
            self.slope = 2.0
        elif isinstance(slope, (int, float)):
            self.slope = float(slope)
        else:
            raise SyntaxError(f'invalid slope type, must be number: {slope}')

        vib_obj = opts.get('vib_obj')
        if vib_obj is None:
            self.vib_obj: VibObjType = {
                'periods': 8,
                'vert_offset': 0,
                'init_up': True,
                'extent': 0.05,
            }
        else:
            # Normalize and validate vib_obj for flexible inputs (e.g., strings)
            self.vib_obj = self._normalize_vib_obj(vib_obj)  # type: ignore

        instr = opts.get('instrumentation', Instrument.Sitar)
        self.instrumentation: Instrument = instr

        articulations_in = opts.get('articulations')
        if articulations_in is None:
            if self.instrumentation == Instrument.Sitar:
                self.articulations: Dict[str, Articulation] = {
                    '0.00': Articulation({'name': 'pluck', 'stroke': 'd'})
                }
            else:
                self.articulations = {}
        else:
            if not isinstance(articulations_in, dict):
                raise SyntaxError(f'invalid articulations type, must be object: {articulations_in}')
            self.articulations = {}
            for k, v in articulations_in.items():
                if not isinstance(v, Articulation):
                    v = Articulation(v)  # type: ignore
                self.articulations[str(k)] = v

        self.num = opts.get('num')
        self.name = opts.get('name')
        self.name = self.name_
        self.ids: List[Callable[[float], float]] = []
        for i in range(14):
            if i == 11:
                self.ids.append(self.id7)
            else:
                self.ids.append(getattr(self, f'id{i}'))
        self.fund_id12 = opts.get('fund_id12')
        self.structured_names = {
            'fixed': 0,
            'bend': {
                'simple': 1,
                'sloped start': 2,
                'sloped end': 3,
                'ladle': 4,
                'reverse ladle': 5,
                'yoyo': 6,
            },
            'krintin': {
                'krintin': 7,
                'krintin slide': 8,
                'krintin slide hammer': 9,
                'spiffy krintin slide hammer': 10,
            },
            'slide': 11,
            'silent': 12,
            'vibrato': 13,
        }
        self.vowel = opts.get('vowel')
        self.vowel_ipa = opts.get('vowel_ipa')
        self.vowel_hindi = opts.get('vowel_hindi')
        self.vowel_eng_trans = opts.get('vowel_eng_trans')
        self.start_consonant = opts.get('start_consonant')
        self.start_consonant_hindi = opts.get('start_consonant_hindi')
        self.start_consonant_ipa = opts.get('start_consonant_ipa')
        self.start_consonant_eng_trans = opts.get('start_consonant_eng_trans')
        self.end_consonant = opts.get('end_consonant')
        self.end_consonant_hindi = opts.get('end_consonant_hindi')
        self.end_consonant_ipa = opts.get('end_consonant_ipa')
        self.end_consonant_eng_trans = opts.get('end_consonant_eng_trans')
        self.group_id = opts.get('group_id')

        automation_in = opts.get('automation')
        if automation_in is not None:
            if isinstance(automation_in, Automation):
                self.automation = automation_in
            else:
                self.automation = Automation(automation_in)
        elif self.id == 12:
            self.automation = None
        else:
            self.automation = Automation()

        if self.start_consonant is not None:
            self.articulations['0.00'] = Articulation({
                'name': 'consonant',
                'stroke': self.start_consonant,
                'hindi': self.start_consonant_hindi,
                'ipa': self.start_consonant_ipa,
            })
        if self.end_consonant is not None:
            self.articulations['1.00'] = Articulation({
                'name': 'consonant',
                'stroke': self.end_consonant,
                'hindi': self.end_consonant_hindi,
                'ipa': self.end_consonant_ipa,
            })

        if self.id < 4:
            self.dur_array = [1]
        elif self.dur_array is None and self.id == 4:
            self.dur_array = [1/3, 2/3]
        elif self.dur_array is None and self.id == 5:
            self.dur_array = [2/3, 1/3]
        elif self.dur_array is None and self.id == 6:
            if len(self.log_freqs) > 1:
                self.dur_array = [1/(len(self.log_freqs)-1)] * (len(self.log_freqs)-1)
            else:
                self.dur_array = []
        elif self.id == 7:
            if self.dur_array is None:
                self.dur_array = [0.2, 0.8]
            starts = get_starts(self.dur_array)
            cond = len(self.log_freqs) > 1 and self.log_freqs[1] >= self.log_freqs[0]
            self.articulations[str(starts[1])] = Articulation({
                'name': 'hammer-on' if cond else 'hammer-off'
            })
        elif self.id == 8:
            if self.dur_array is None:
                self.dur_array = [1/3,1/3,1/3]
            starts = get_starts(self.dur_array)
            self.articulations[str(starts[1])] = Articulation({'name': 'hammer-off'})
            self.articulations[str(starts[2])] = Articulation({'name': 'slide'})
        elif self.id == 9:
            if self.dur_array is None:
                self.dur_array = [0.25,0.25,0.25,0.25]
            starts = get_starts(self.dur_array)
            self.articulations[str(starts[1])] = Articulation({'name': 'hammer-off'})
            self.articulations[str(starts[2])] = Articulation({'name': 'slide'})
            self.articulations[str(starts[3])] = Articulation({'name': 'hammer-on'})
        elif self.id == 10:
            if self.dur_array is None:
                self.dur_array = [1/6]*6
            starts = get_starts(self.dur_array)
            self.articulations[str(starts[1])] = Articulation({'name': 'slide'})
            self.articulations[str(starts[2])] = Articulation({'name': 'hammer-on'})
            self.articulations[str(starts[3])] = Articulation({'name': 'hammer-off'})
            self.articulations[str(starts[4])] = Articulation({'name': 'slide'})
            self.articulations[str(starts[5])] = Articulation({'name': 'hammer-on'})
        elif self.id == 11:
            if self.dur_array is None or len(self.dur_array) == 1:
                self.dur_array = [0.5,0.5]
            starts = get_starts(self.dur_array)
            self.articulations[str(starts[1])] = Articulation({'name': 'slide'})

        if self.dur_array:
            i = 0
            while i < len(self.dur_array):
                if self.dur_array[i] == 0:
                    print('removing zero dur')
                    self.dur_array.pop(i)
                    if i+1 < len(self.pitches):
                        self.pitches.pop(i+1)
                else:
                    i += 1

        if self.instrumentation in (Instrument.Vocal_M, Instrument.Vocal_F):
            for k in list(self.articulations.keys()):
                if self.articulations[k].name == 'pluck':
                    del self.articulations[k]

        self.c_ipas = ['k', 'kʰ', 'g', 'gʱ', 'ŋ', 'c', 'cʰ', 'ɟ', 'ɟʱ', 'ɲ', 'ʈ',
                       'ʈʰ', 'ɖ', 'ɖʱ', 'n', 't', 'tʰ', 'd', 'dʱ', 'n̪', 'p', 'pʰ', 'b', 'bʱ',
                       'm', 'j', 'r', 'l', 'v', 'ʃ', 'ʂ', 's', 'h']
        self.c_isos = ['ka', 'kha', 'ga', 'gha', 'ṅa', 'ca', 'cha', 'ja', 'jha', 'ña', 'ṭa',
                       'ṭha', 'ḍa', 'ḍha', 'na', 'ta', 'tha', 'da', 'dha', 'na', 'pa', 'pha',
                       'ba', 'bha', 'ma', 'ya', 'ra', 'la', 'va', 'śa', 'ṣa', 'sa', 'ha']
        self.c_hindis = ['क', 'ख', 'ग', 'घ', 'ङ', 'च', 'छ', 'ज', 'झ', 'ञ', 'ट',
                          'ठ', 'ड', 'ढ', 'न', 'त', 'थ', 'द', 'ध', 'न', 'प', 'फ़', 'ब', 'भ', 'म', 'य',
                          'र', 'ल', 'व', 'श', 'ष', 'स', 'ह']
        self.c_eng_trans = ['k', 'kh', 'g', 'gh', 'ṅ', 'c', 'ch', 'j', 'jh', 'ñ', 'ṭ',
                             'ṭh', 'ḍ', 'ḍh', 'n', 't', 'th', 'd', 'dh', 'n', 'p', 'ph', 'b', 'bh',
                             'm', 'y', 'r', 'l', 'v', 'ś', 'ṣ', 's', 'h']
        self.v_ipas = ['ə', 'aː', 'ɪ', 'iː', 'ʊ', 'uː', 'eː', 'ɛː', 'oː', 'ɔː', '_']
        self.v_isos = ['a', 'ā', 'i', 'ī', 'u', 'ū', 'ē', 'ai', 'ō', 'au', '_']
        self.v_hindis = ['अ', 'आ', 'इ', 'ई', 'उ', 'ऊ', 'ए', 'ऐ', 'ओ', 'औ', '_']
        self.v_eng_trans = ['a', 'ā', 'i', 'ī', 'u', 'ū', 'ē', 'ai', 'ō', 'au', '_']

        self.unique_id = opts.get('unique_id') or str(uuid.uuid4())
        self.convert_c_iso_to_hindi_and_ipa()

        for k in list(self.articulations.keys()):
            if k == '0':
                self.articulations['0.00'] = self.articulations[k]
                del self.articulations[k]

        self.tags = opts.get('tags', [])

        self.start_time: Optional[float] = opts.get('start_time')
        self.phrase_idx: Optional[int] = None

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {
            'id', 'pitches', 'dur_tot', 'dur_array', 'slope', 'vib_obj', 'instrumentation',
            'articulations', 'num', 'name', 'fund_id12', 'vowel', 'vowel_ipa', 'vowel_hindi',
            'vowel_eng_trans', 'start_consonant', 'start_consonant_hindi', 'start_consonant_ipa',
            'start_consonant_eng_trans', 'end_consonant', 'end_consonant_hindi', 'end_consonant_ipa',
            'end_consonant_eng_trans', 'group_id', 'automation', 'unique_id', 'tags', 'start_time',
            'phrase_idx'
        }
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names with helpful suggestions
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'type':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'id'?")
                elif key == 'duration':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'dur_tot'?")
                elif key == 'instrument':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'instrumentation'?")
                elif key == 'duration_array':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'dur_array'?")
                elif key == 'vibrato_obj':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'vib_obj'?")
                elif key == 'fundamental_id12':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'fund_id12'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        self._validate_parameter_types(opts)
        self._validate_parameter_values(opts)
    
    def _validate_parameter_types(self, opts: dict) -> None:
        """Validate that all parameters have correct types.
        Note: Some parameters (id, pitches, dur_tot, slope, articulations) are validated 
        by the original constructor logic which throws SyntaxError, so we skip them here."""
        
        # Skip parameters that are already validated by original constructor logic:
        # - id, pitches, dur_tot, slope, articulations (validated with SyntaxError)
        
        if 'dur_array' in opts and opts['dur_array'] is not None:
            if not isinstance(opts['dur_array'], list):
                raise TypeError(f"Parameter 'dur_array' must be a list, got {type(opts['dur_array']).__name__}")
            if not all(isinstance(d, (int, float)) for d in opts['dur_array']):
                raise TypeError("All items in 'dur_array' must be numbers")
        
        if 'vib_obj' in opts and opts['vib_obj'] is not None:
            if not isinstance(opts['vib_obj'], dict):
                raise TypeError(f"Parameter 'vib_obj' must be a dict, got {type(opts['vib_obj']).__name__}")
            self._validate_vib_obj_structure(opts['vib_obj'])
        
        if 'instrumentation' in opts and not isinstance(opts['instrumentation'], Instrument):
            raise TypeError(f"Parameter 'instrumentation' must be an Instrument enum, got {type(opts['instrumentation']).__name__}")
        
        if 'start_time' in opts and opts['start_time'] is not None:
            if not isinstance(opts['start_time'], (int, float)):
                raise TypeError(f"Parameter 'start_time' must be a number, got {type(opts['start_time']).__name__}")
        
        # Validate string parameters
        string_params = ['name', 'vowel', 'vowel_ipa', 'vowel_hindi', 'vowel_eng_trans',
                        'start_consonant', 'start_consonant_hindi', 'start_consonant_ipa',
                        'start_consonant_eng_trans', 'end_consonant', 'end_consonant_hindi',
                        'end_consonant_ipa', 'end_consonant_eng_trans', 'unique_id']
        
        for param in string_params:
            if param in opts and opts[param] is not None and not isinstance(opts[param], str):
                raise TypeError(f"Parameter '{param}' must be a string, got {type(opts[param]).__name__}")
        
        if 'tags' in opts and not isinstance(opts['tags'], list):
            raise TypeError(f"Parameter 'tags' must be a list, got {type(opts['tags']).__name__}")
    
    def _validate_parameter_values(self, opts: dict) -> None:
        """Validate that parameter values are in valid ranges.
        Note: Some parameters (id, dur_tot, slope) are validated by original constructor,
        so we only do additional range checks here."""
        
        # Additional range validation for parameters already type-checked by original constructor
        if 'id' in opts and isinstance(opts['id'], int):
            if not 0 <= opts['id'] <= 13:
                raise ValueError(f"Parameter 'id' must be between 0-13 (trajectory types), got {opts['id']}")
        
        if 'dur_tot' in opts and isinstance(opts['dur_tot'], (int, float)):
            if opts['dur_tot'] <= 0:
                raise ValueError(f"Parameter 'dur_tot' must be positive, got {opts['dur_tot']}")
        
        if 'dur_array' in opts and opts['dur_array'] is not None:
            dur_array = opts['dur_array']
            if any(d < 0 for d in dur_array):
                raise ValueError("All values in 'dur_array' must be non-negative")
            if len(dur_array) > 0 and sum(dur_array) == 0:
                raise ValueError("'dur_array' cannot have all zero values")
        
        if 'slope' in opts and isinstance(opts['slope'], (int, float)):
            if opts['slope'] <= 0:
                raise ValueError(f"Parameter 'slope' must be positive, got {opts['slope']}")
        
        if 'start_time' in opts and opts['start_time'] is not None:
            if opts['start_time'] < 0:
                raise ValueError(f"Parameter 'start_time' must be non-negative, got {opts['start_time']}")
        
        # Validate vocal parameters are only used with vocal instruments
        vocal_params = ['vowel', 'vowel_ipa', 'vowel_hindi', 'vowel_eng_trans',
                       'start_consonant', 'start_consonant_hindi', 'start_consonant_ipa',
                       'start_consonant_eng_trans', 'end_consonant', 'end_consonant_hindi',
                       'end_consonant_ipa', 'end_consonant_eng_trans']
        
        has_vocal_params = any(param in opts and opts[param] is not None for param in vocal_params)
        instrumentation = opts.get('instrumentation', Instrument.Sitar)
        
        if has_vocal_params and instrumentation not in (Instrument.Vocal_M, Instrument.Vocal_F):
            import warnings
            warnings.warn(f"Vocal parameters provided but instrumentation is {instrumentation.name}. "
                         "Vocal parameters are typically used with Vocal_M or Vocal_F instruments.", UserWarning)
    
    def _validate_vib_obj_structure(self, vib_obj: dict) -> None:
        """Validate vib_obj has correct structure, allowing lenient input types.

        Accepts numeric strings and floats that can be coerced to required types,
        but does not mutate the provided dict. Actual coercion happens in
        _normalize_vib_obj.
        """
        allowed_keys = {'periods', 'vert_offset', 'init_up', 'extent'}
        provided_keys = set(vib_obj.keys())
        invalid_keys = provided_keys - allowed_keys

        if invalid_keys:
            raise ValueError(
                f"vib_obj contains invalid keys: {sorted(invalid_keys)}. "
                f"Allowed keys: {sorted(allowed_keys)}"
            )

        # Validate types and values (lenient)
        if 'periods' in vib_obj:
            p = vib_obj['periods']
            if isinstance(p, int):
                if p <= 0:
                    raise ValueError("vib_obj['periods'] must be positive")
            elif isinstance(p, float):
                if p <= 0:
                    raise ValueError("vib_obj['periods'] must be positive")
            elif isinstance(p, str):
                try:
                    pf = float(p.strip())
                except Exception as e:
                    raise TypeError("vib_obj['periods'] must be an integer or numeric string") from e
                if pf <= 0:
                    raise ValueError("vib_obj['periods'] must be positive")
            else:
                raise TypeError("vib_obj['periods'] must be a number")

        for key in ['vert_offset', 'extent']:
            if key in vib_obj:
                v = vib_obj[key]
                if isinstance(v, (int, float)):
                    pass
                elif isinstance(v, str):
                    try:
                        float(v.strip())
                    except Exception as e:
                        raise TypeError(f"vib_obj['{key}'] must be a number or numeric string") from e
                else:
                    raise TypeError(f"vib_obj['{key}'] must be a number")

        if 'extent' in vib_obj:
            try:
                ext_val = float(vib_obj['extent'])
            except Exception:
                # If not coercible, earlier checks will have raised
                ext_val = 0.0
            if ext_val <= 0:
                raise ValueError("vib_obj['extent'] must be positive")

        if 'init_up' in vib_obj:
            iu = vib_obj['init_up']
            if isinstance(iu, bool):
                pass
            elif isinstance(iu, (int, float)):
                if iu not in (0, 1):
                    raise TypeError("vib_obj['init_up'] must be boolean-like (0/1)")
            elif isinstance(iu, str):
                if iu.strip().lower() not in {'true', 'false', '0', '1'}:
                    raise TypeError("vib_obj['init_up'] must be 'true'/'false' or '0'/'1'")
            else:
                raise TypeError("vib_obj['init_up'] must be a boolean or boolean-like string")

    def _normalize_vib_obj(self, vib_obj: dict) -> VibObjType:
        """Return a normalized VibObjType with correct Python types.

        - periods: int (>0)
        - vert_offset: float
        - extent: float (>0)
        - init_up: bool
        """
        # Start from defaults
        norm: VibObjType = {
            'periods': 8,
            'vert_offset': 0.0,
            'init_up': True,
            'extent': 0.05,
        }

        # Validate structure leniently first
        self._validate_vib_obj_structure(vib_obj)

        # Coerce values
        if 'periods' in vib_obj:
            p = vib_obj['periods']
            if isinstance(p, (int, float)):
                norm['periods'] = int(p)
            elif isinstance(p, str):
                norm['periods'] = int(float(p.strip()))

        if 'vert_offset' in vib_obj:
            v = vib_obj['vert_offset']
            if isinstance(v, (int, float)):
                norm['vert_offset'] = float(v)
            elif isinstance(v, str):
                norm['vert_offset'] = float(v.strip())

        if 'extent' in vib_obj:
            e = vib_obj['extent']
            if isinstance(e, (int, float)):
                norm['extent'] = float(e)
            elif isinstance(e, str):
                norm['extent'] = float(e.strip())

        if 'init_up' in vib_obj:
            iu = vib_obj['init_up']
            if isinstance(iu, bool):
                norm['init_up'] = iu
            elif isinstance(iu, (int, float)):
                norm['init_up'] = bool(int(iu))
            elif isinstance(iu, str):
                sval = iu.strip().lower()
                if sval in {'true', '1'}:
                    norm['init_up'] = True
                elif sval in {'false', '0'}:
                    norm['init_up'] = False
                else:
                    # Should not happen due to validation above
                    raise TypeError("vib_obj['init_up'] string must be 'true'/'false' or '0'/'1'")

        # Final sanity checks
        if norm['periods'] <= 0:
            raise ValueError("vib_obj['periods'] must be positive after normalization")
        if norm['extent'] <= 0:
            raise ValueError("vib_obj['extent'] must be positive after normalization")

        return norm

    # ------------------------------- properties -----------------------------
    @property
    def freqs(self) -> List[float]:
        return [p.frequency for p in self.pitches]

    @property
    def log_freqs(self) -> List[float]:
        return [math.log2(p.frequency) for p in self.pitches]

    @property
    def sloped(self) -> bool:
        return self.id in (2,3,4,5)

    @property
    def min_freq(self) -> float:
        return min(self.freqs)

    @property
    def max_freq(self) -> float:
        return max(self.freqs)

    @property
    def min_log_freq(self) -> float:
        return min(self.log_freqs)

    @property
    def max_log_freq(self) -> float:
        return max(self.log_freqs)

    @property
    def end_time(self) -> Optional[float]:
        if self.start_time is None:
            return None
        return self.start_time + self.dur_tot

    @property
    def name_(self) -> str:
        return self.names[self.id]

    # ------------------------------- utils -----------------------------
    def update_fundamental(self, fundamental: float) -> None:
        for p in self.pitches:
            p.fundamental = fundamental

    # ------------------------------- computation -----------------------
    def compute(self, x: float, log_scale: bool = False) -> float:
        val = self.ids[self.id](x)
        return math.log2(val) if log_scale else val

    def id0(self, x: float, lf: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        return 2 ** log_freqs[0]

    def id1(self, x: float, lf: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        pi_x = (math.cos(math.pi * (x + 1)) / 2) + 0.5
        diff = log_freqs[1] - log_freqs[0]
        return 2 ** (pi_x * diff + log_freqs[0])

    def id2(self, x: float, lf: Optional[List[float]] = None, sl: Optional[float] = None) -> float:
        log_freqs = lf or self.log_freqs
        slope = sl if sl is not None else self.slope
        a = log_freqs[0]
        b = log_freqs[1]
        log_freq_out = (a - b) * ((1 - x) ** slope) + b
        return 2 ** log_freq_out

    def id3(self, x: float, lf: Optional[List[float]] = None, sl: Optional[float] = None) -> float:
        log_freqs = lf or self.log_freqs
        slope = sl if sl is not None else self.slope
        a = log_freqs[0]
        b = log_freqs[1]
        log_freq_out = (b - a) * (x ** slope) + a
        return 2 ** log_freq_out

    def id4(self, x: float, lf: Optional[List[float]] = None, sl: Optional[float] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        slope = sl if sl is not None else self.slope
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [1/3,2/3]
        bend0 = lambda x: self.id2(x, log_freqs[:2], slope)
        bend1 = lambda x: self.id1(x, log_freqs[1:3])
        out0 = lambda x: bend0(x / dur_array[0])
        out1 = lambda x: bend1((x - dur_array[0]) / dur_array[1])
        return out0(x) if x < dur_array[0] else out1(x)

    def id5(self, x: float, lf: Optional[List[float]] = None, sl: Optional[float] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        slope = sl if sl is not None else self.slope
        dur_array = da if da is not None else self.dur_array
        dur_array = dur_array or [1/3,2/3]
        bend0 = lambda x: self.id1(x, log_freqs[:2])
        bend1 = lambda x: self.id3(x, log_freqs[1:3], slope)
        out0 = lambda x: bend0(x / dur_array[0])
        out1 = lambda x: bend1((x - dur_array[0]) / dur_array[1])
        return out0(x) if x < dur_array[0] else out1(x)

    def id6(self, x: float, lf: Optional[List[float]] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [1/(len(log_freqs)-1)] * (len(log_freqs)-1)
        
        # Get segment start points
        starts = get_starts(dur_array)
        
        # Find the correct segment index using proper boundary logic
        # This matches the TypeScript findLastIndex behavior
        index = -1
        for i in range(len(starts)):
            if x >= starts[i]:
                # Check if this is the last segment or if x is before the next segment start
                if i == len(starts) - 1 or x < starts[i + 1]:
                    index = i
                    break
        
        if index == -1:
            # Fallback for edge cases (x < 0)
            index = 0
        
        # Create the interpolation function for this segment
        bend = lambda y: self.id1(y, log_freqs[index:index+2])
        
        # Calculate the relative position within this segment
        dur_sum = sum(dur_array[:index])
        relative_x = (x - dur_sum) / dur_array[index]
        
        # Ensure relative_x is within [0, 1] bounds
        relative_x = max(0.0, min(1.0, relative_x))
        
        return bend(relative_x)

    def id7(self, x: float, lf: Optional[List[float]] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [0.5,0.5]
        out = log_freqs[0] if x < dur_array[0] else log_freqs[1]
        return 2 ** out

    def id8(self, x: float, lf: Optional[List[float]] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [1/3,1/3,1/3]
        starts = get_starts(dur_array)
        index = 0
        for i,s in enumerate(starts):
            if x >= s:
                index = i
        return 2 ** log_freqs[index]

    def id9(self, x: float, lf: Optional[List[float]] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [0.25,0.25,0.25,0.25]
        starts = get_starts(dur_array)
        index = 0
        for i,s in enumerate(starts):
            if x >= s:
                index = i
        return 2 ** log_freqs[index]

    def id10(self, x: float, lf: Optional[List[float]] = None, da: Optional[List[float]] = None) -> float:
        log_freqs = lf or self.log_freqs
        dur_array = da if da is not None else self.dur_array
        if dur_array is None:
            dur_array = [i/6 for i in range(6)]
        starts = get_starts(dur_array)
        index = 0
        for i,s in enumerate(starts):
            if x >= s:
                index = i
        return 2 ** log_freqs[index]

    def id12(self, x: float) -> float:
        return float(self.fund_id12)

    def id13(self, x: float) -> float:
        periods = self.vib_obj['periods']
        vert_offset = self.vib_obj['vert_offset']
        init_up = self.vib_obj['init_up']
        extent = self.vib_obj['extent']
        if abs(vert_offset) > extent / 2:
            vert_offset = math.copysign(extent/2, vert_offset)
        out = math.cos(x * 2 * math.pi * periods + int(init_up) * math.pi)
        if x < 1/(2*periods):
            start = self.log_freqs[0]
            end = math.log2(self.id13(1/(2*periods)))
            middle = (end + start)/2
            ext = abs(end - start)/2
            out = out*ext + middle
            return 2 ** out
        elif x > 1 - 1/(2*periods):
            start = math.log2(self.id13(1 - 1/(2*periods)))
            end = self.log_freqs[0]
            middle = (end + start)/2
            ext = abs(end - start)/2
            out = out*ext + middle
            return 2 ** out
        else:
            return 2 ** (out * extent/2 + vert_offset + self.log_freqs[0])

    # ---------------- consonant/vowel helpers -----------------------
    def remove_consonant(self, start: bool = True) -> None:
        if start:
            self.start_consonant = None
            self.start_consonant_hindi = None
            self.start_consonant_ipa = None
            self.start_consonant_eng_trans = None
            art = self.articulations.get('0.00')
            if art and art.name == 'consonant':
                del self.articulations['0.00']
        else:
            self.end_consonant = None
            self.end_consonant_hindi = None
            self.end_consonant_ipa = None
            self.end_consonant_eng_trans = None
            art = self.articulations.get('1.00')
            if art and art.name == 'consonant':
                del self.articulations['1.00']

    def add_consonant(self, consonant: str, start: bool = True) -> None:
        idx = self.c_isos.index(consonant) if consonant in self.c_isos else -1
        hindi = self.c_hindis[idx] if idx != -1 else None
        ipa = self.c_ipas[idx] if idx != -1 else None
        eng = self.c_eng_trans[idx] if idx != -1 else None
        art = Articulation({'name': 'consonant', 'stroke': consonant, 'hindi': hindi, 'ipa': ipa, 'eng_trans': eng})
        if start:
            self.start_consonant = consonant
            self.start_consonant_hindi = hindi
            self.start_consonant_ipa = ipa
            self.start_consonant_eng_trans = eng
            self.articulations['0.00'] = art
        else:
            self.end_consonant = consonant
            self.end_consonant_hindi = hindi
            self.end_consonant_ipa = ipa
            self.end_consonant_eng_trans = eng
            self.articulations['1.00'] = art

    def change_consonant(self, consonant: str, start: bool = True) -> None:
        idx = self.c_isos.index(consonant) if consonant in self.c_isos else -1
        hindi = self.c_hindis[idx] if idx != -1 else None
        ipa = self.c_ipas[idx] if idx != -1 else None
        eng = self.c_eng_trans[idx] if idx != -1 else None
        if start:
            self.start_consonant = consonant
            self.start_consonant_hindi = hindi
            self.start_consonant_ipa = ipa
            self.start_consonant_eng_trans = eng
            art = self.articulations['0.00']
            art.stroke = consonant
            art.hindi = hindi
            art.ipa = ipa
            art.eng_trans = eng
        else:
            self.end_consonant = consonant
            self.end_consonant_hindi = hindi
            self.end_consonant_ipa = ipa
            self.end_consonant_eng_trans = eng
            art = self.articulations['1.00']
            art.stroke = consonant
            art.hindi = hindi
            art.ipa = ipa
            art.eng_trans = eng

    def durations_of_fixed_pitches(self, opts: Optional[Dict] = None) -> Dict:
        output_type = 'pitchNumber'
        if opts:
            output_type = opts.get('output_type', 'pitchNumber')
        pitch_durs: Dict = {}
        id_str = str(self.id)
        if id_str in ('0','13'):
            pitch_durs[self.pitches[0].numbered_pitch] = self.dur_tot
        elif id_str in ('1','2','3'):
            if self.pitches[0].numbered_pitch == self.pitches[1].numbered_pitch:
                pitch_durs[self.pitches[0].numbered_pitch] = self.dur_tot
        elif id_str in ('4','5'):
            p0 = self.pitches[0].numbered_pitch
            p1 = self.pitches[1].numbered_pitch
            p2 = self.pitches[2].numbered_pitch
            if p0 == p1:
                pitch_durs[p0] = self.dur_tot * self.dur_array[0]
            elif p1 == p2:
                pitch_durs[p1] = pitch_durs.get(p1,0) + self.dur_tot * self.dur_array[1]
        elif id_str == '6':
            last_num = None
            for i,p in enumerate(self.pitches):
                num = p.numbered_pitch
                if num == last_num:
                    pitch_durs[num] = pitch_durs.get(num,0) + self.dur_tot * self.dur_array[i-1]
                last_num = num
        elif id_str in ('7','8','9','10','11'):
            for i,p in enumerate(self.pitches):
                if i < len(self.dur_array) and self.dur_array[i] is not None:
                    num = p.numbered_pitch
                    pitch_durs[num] = pitch_durs.get(num,0) + self.dur_tot * self.dur_array[i]
        if output_type == 'pitchNumber':
            return pitch_durs
        elif output_type == 'chroma':
            alt = {}
            for p,v in pitch_durs.items():
                c = Pitch.pitch_number_to_chroma(int(p))
                alt[c] = v
            return alt
        elif output_type == 'scaleDegree':
            alt = {}
            for p,v in pitch_durs.items():
                c = Pitch.pitch_number_to_chroma(int(p))
                sd = Pitch.chroma_to_scale_degree(c)[0]
                alt[sd] = v
            return alt
        elif output_type == 'sargamLetter':
            alt = {}
            for p,v in pitch_durs.items():
                s = Pitch.from_pitch_number(int(p)).sargam_letter
                alt[s] = v
            return alt
        else:
            raise Exception('outputType not recognized')

    def convert_c_iso_to_hindi_and_ipa(self) -> None:
        for art in self.articulations.values():
            if art.name == 'consonant':
                if not isinstance(art.stroke, str):
                    raise Exception('stroke is not a string')
                c_iso = art.stroke
                if c_iso in self.c_isos:
                    idx = self.c_isos.index(c_iso)
                    art.hindi = getattr(art, 'hindi', None) or self.c_hindis[idx]
                    art.ipa = getattr(art, 'ipa', None) or self.c_ipas[idx]
                    art.eng_trans = getattr(art, 'eng_trans', None) or self.c_eng_trans[idx]
                else:
                    if not hasattr(art, 'hindi'):
                        art.hindi = None
                    if not hasattr(art, 'ipa'):
                        art.ipa = None
                    if not hasattr(art, 'eng_trans'):
                        art.eng_trans = None
        if self.start_consonant is not None:
            c_iso = self.start_consonant
            if c_iso in self.c_isos:
                idx = self.c_isos.index(c_iso)
                self.start_consonant_hindi = getattr(self, 'start_consonant_hindi', None) or self.c_hindis[idx]
                self.start_consonant_ipa = getattr(self, 'start_consonant_ipa', None) or self.c_ipas[idx]
                self.start_consonant_eng_trans = getattr(self, 'start_consonant_eng_trans', None) or self.c_eng_trans[idx]
        if self.end_consonant is not None:
            c_iso = self.end_consonant
            if c_iso in self.c_isos:
                idx = self.c_isos.index(c_iso)
                self.end_consonant_hindi = getattr(self, 'end_consonant_hindi', None) or self.c_hindis[idx]
                self.end_consonant_ipa = getattr(self, 'end_consonant_ipa', None) or self.c_ipas[idx]
                self.end_consonant_eng_trans = getattr(self, 'end_consonant_eng_trans', None) or self.c_eng_trans[idx]
        if self.vowel is not None:
            v_iso = self.vowel
            if v_iso in self.v_isos:
                idx = self.v_isos.index(v_iso)
                self.vowel_hindi = getattr(self, 'vowel_hindi', None) or self.v_hindis[idx]
                self.vowel_ipa = getattr(self, 'vowel_ipa', None) or self.v_ipas[idx]
                self.vowel_eng_trans = getattr(self, 'vowel_eng_trans', None) or self.v_eng_trans[idx]

    def update_vowel(self, v_iso: str) -> None:
        if v_iso in self.v_isos:
            idx = self.v_isos.index(v_iso)
            self.vowel = v_iso
            self.vowel_hindi = self.v_hindis[idx]
            self.vowel_ipa = self.v_ipas[idx]
            self.vowel_eng_trans = self.v_eng_trans[idx]
        else:
            self.vowel = v_iso
            self.vowel_hindi = None
            self.vowel_ipa = None
            self.vowel_eng_trans = None

    def to_json(self) -> Dict:
        data = {
            'id': self.id,
            'pitches': [p.to_json() for p in self.pitches],
            'durTot': self.dur_tot,
            'durArray': self.dur_array,
            'slope': self.slope,
            'articulations': {k: a.to_json() for k, a in self.articulations.items()},
            'startTime': self.start_time,
            'num': self.num,
            'name': self.name,
            'fundID12': self.fund_id12,
            'vibObj': self.vib_obj,
            'instrumentation': self.instrumentation.value if isinstance(self.instrumentation, Instrument) else self.instrumentation,
            'vowel': self.vowel,
            'startConsonant': self.start_consonant,
            'startConsonantHindi': self.start_consonant_hindi,
            'startConsonantIpa': self.start_consonant_ipa,
            'startConsonantEngTrans': self.start_consonant_eng_trans,
            'endConsonant': self.end_consonant,
            'endConsonantHindi': self.end_consonant_hindi,
            'endConsonantIpa': self.end_consonant_ipa,
            'endConsonantEngTrans': self.end_consonant_eng_trans,
            'groupId': self.group_id,
            'automation': self.automation.to_json() if self.automation else None,
            'uniqueId': self.unique_id,
            'tags': self.tags,
        }
        # drop None values so they serialize as undefined (omitted) rather than null
        return {k: v for k, v in data.items() if v is not None}

    @staticmethod
    def from_json(obj: Dict) -> 'Trajectory':
        opts = humps.decamelize(obj)
        pitches = [Pitch.from_json(p) for p in opts.get('pitches', [])]
        arts = {}
        for k,v in opts.get('articulations', {}).items():
            if v is not None:
                arts[k] = Articulation.from_json(v)
        automation = opts.get('automation')
        instr = opts.get('instrumentation')
        if isinstance(instr, str):
            try:
                opts['instrumentation'] = Instrument(instr)
            except ValueError:
                opts['instrumentation'] = Instrument.Sitar
        opts['pitches'] = pitches
        opts['articulations'] = arts
        opts['automation'] = automation
        return Trajectory(opts)

    @staticmethod
    def names() -> List[str]:
        return Trajectory().names
